import numpy as np
from numpy.testing import assert_array_equal
from nose.tools import (assert_equal, assert_raises, raises, assert_true,
                        assert_false)
from nose.tools import with_setup

try:
    from nose.tools import assert_is, assert_is_instance
except ImportError:
    from landlab.testing.tools import assert_is, assert_is_instance

from landlab import RasterModelGrid
from landlab import BAD_INDEX_VALUE as X


def setup_grid():
    """
    These tests use a grid that 4x5 nodes::

     15------16------17------18------19
      |       |       |       |       |
      |       |       |       |       |
      |       |       |       |       |
     10------11------12------13------14
      |       |       |       |       |
      |       |       |       |       |   
      |       |       |       |       |
      5-------6-------7-------8-------9
      |       |       |       |       |
      |       |       |       |       |
      |       |       |       |       |
      0-------1-------2-------3-------4
    """
    from landlab import RasterModelGrid
    globals().update({
        'rmg': RasterModelGrid(4, 5, 1.)
    })


def test_init_with_kwds_classic():
    grid = RasterModelGrid(num_rows=4, num_cols=5, dx=1.)

    assert_equal(grid.number_of_node_rows, 4)
    assert_equal(grid.number_of_node_columns, 5)
    assert_equal(grid.node_spacing, 1)

    grid = RasterModelGrid(3, 7, 2)

    assert_equal(grid.number_of_node_rows, 3)
    assert_equal(grid.number_of_node_columns, 7)
    assert_equal(grid.node_spacing, 2.)


def test_init_new_style():
    grid = RasterModelGrid((4, 5), spacing=2)

    assert_equal(grid.number_of_node_rows, 4)
    assert_equal(grid.number_of_node_columns, 5)
    assert_equal(grid.node_spacing, 2.)

    grid = RasterModelGrid((4, 5))

    assert_equal(grid.number_of_node_rows, 4)
    assert_equal(grid.number_of_node_columns, 5)
    assert_equal(grid.node_spacing, 1.)


def test_spacing_is_float():
    grid = RasterModelGrid((4, 5))
    assert_equal(grid.node_spacing, 1.)
    assert_is_instance(grid.node_spacing, float)

    grid = RasterModelGrid((4, 5), spacing=2)
    assert_equal(grid.node_spacing, 2.)
    assert_is_instance(grid.node_spacing, float)


@with_setup(setup_grid)
def test_grid_dimensions():
    assert_equal(rmg.get_grid_ydimension(), rmg.number_of_node_rows - 1)
    assert_equal(rmg.get_grid_xdimension(), rmg.number_of_node_columns - 1)


def test_grid_dimensions_non_unit_spacing():
    rmg = RasterModelGrid((4, 5), spacing=2.)
    assert_equal(rmg.get_grid_ydimension(), 6.)
    assert_equal(rmg.get_grid_xdimension(), 8.)


@with_setup(setup_grid)
def test_nodes_around_point():
    surrounding_ids = rmg.get_nodes_around_point(2.1, 1.1)
    assert_array_equal(surrounding_ids, np.array([7, 12, 13, 8]))

    surrounding_ids = rmg.get_nodes_around_point(2.1, .9)
    assert_array_equal(surrounding_ids, np.array([2, 7, 8, 3]))


@with_setup(setup_grid)
def test_neighbor_list_with_scalar_arg():
    assert_array_equal(rmg.get_neighbor_list(6), np.array([7, 11, 5, 1]))
    assert_array_equal(rmg.get_neighbor_list(-1), np.array([X, X, X, X]))
    assert_array_equal(rmg.get_neighbor_list(-2), np.array([X, X, X, 13]))


@with_setup(setup_grid)
def test_neighbor_list_with_array_arg():
    assert_array_equal(rmg.get_neighbor_list([6, -1]),
                       np.array([[7, 11, 5, 1], [X, X, X, X]]))


@with_setup(setup_grid)
def test_neighbor_list_with_no_args():
    expected = np.array([
        [ X,  X,  X,  X], [ X,  6,  X,  X], [ X,  7,  X,  X], [ X,  8,  X,  X],
            [ X,  X,  X,  X],
        [ 6,  X,  X,  X], [ 7, 11,  5,  1], [ 8, 12,  6,  2], [ 9, 13,  7,  3],
            [ X,  X,  8,  X],
        [11,  X,  X,  X], [12, 16, 10,  6], [13, 17, 11,  7], [14, 18, 12,  8],
            [ X,  X, 13,  X],
        [ X,  X,  X,  X], [ X,  X,  X, 11], [ X,  X,  X, 12], [ X,  X,  X, 13],
            [ X,  X,  X,  X]])

    assert_array_equal(rmg.get_neighbor_list(), expected)


@with_setup(setup_grid)
def test_node_x():
    assert_array_equal(rmg.node_x, np.array([0., 1., 2., 3., 4.,
                                             0., 1., 2., 3., 4.,
                                             0., 1., 2., 3., 4.,
                                             0., 1., 2., 3., 4.]))

@with_setup(setup_grid)
def test_node_y():
    assert_array_equal(rmg.node_y, np.array([0., 0., 0., 0., 0.,
                                             1., 1., 1., 1., 1.,
                                             2., 2., 2., 2., 2.,
                                             3., 3., 3., 3., 3.]))


@with_setup(setup_grid)
@raises(ValueError)
def test_node_x_is_immutable():
    rmg.node_x[0] = 0


@with_setup(setup_grid)
@raises(ValueError)
def test_node_y_is_immutable():
    rmg.node_y[0] = 0


@with_setup(setup_grid)
def test_node_axis_coordinates():
    assert_is(rmg.node_axis_coordinates(axis=0).base, rmg.node_y.base)
    assert_is(rmg.node_axis_coordinates(axis=1).base, rmg.node_x.base)
    assert_is(rmg.node_axis_coordinates(axis=-1).base, rmg.node_x.base)
    assert_is(rmg.node_axis_coordinates(axis=-2).base, rmg.node_y.base)


@with_setup(setup_grid)
def test_diagonal_list():
    assert_array_equal(rmg.get_diagonal_list(6), np.array([12, 10, 0, 2]))
    assert_array_equal(rmg.get_diagonal_list(-1), np.array([X, X, 13, X]))
    assert_array_equal(rmg.get_diagonal_list([6, -1]),
                       np.array([[12, 10, 0, 2], [X, X, 13, X]]))
    assert_array_equal(
        rmg.get_diagonal_list(),
        np.array([[6, X, X, X], [7, 5, X, X], [8, 6, X, X],
                    [9, 7, X, X], [X, 8, X, X],
                  [11, X, X, 1], [12, 10,  0,  2], [13, 11,  1,  3],
                    [14, 12,  2,  4], [X, 13, 3, X],
                  [16, X, X, 6], [17, 15,  5,  7], [18, 16,  6,  8],
                    [19, 17,  7,  9], [X, 18, 8, X],
                  [X, X, X, 11], [X, X, 10, 12], [X, X, 11, 13],
                    [X, X, 12, 14], [X, X, 13, X]]))


@with_setup(setup_grid)
def test_diagonal_list_boundary():
    assert_array_equal(rmg.get_diagonal_list(0), np.array([6, X, X, X]))


@with_setup(setup_grid)
def test_is_interior():
    for cell_id in [0, 1, 2, 3, 4, 5, 9, 10, 14, 15, 16, 17, 18, 19]:
        assert_false(rmg.is_interior(cell_id))
    for cell_id in [6, 7, 8, 11, 12, 13]:
        assert_true(rmg.is_interior(cell_id))


@with_setup(setup_grid)
def test_get_interior_cells():
    assert_array_equal(rmg.get_core_cell_node_ids(),
                       np.array([6, 7, 8, 11, 12, 13]))


@with_setup(setup_grid)
def test_active_links():
    assert_equal(rmg.number_of_active_links, 17)
    assert_array_equal(rmg.active_links,
                       np.array([1, 2, 3, 6, 7, 8, 11, 12, 13,
                                 19, 20, 21, 22, 23, 24, 25, 26]))


@with_setup(setup_grid)
def test_active_link_fromnode():
    assert_array_equal(rmg.activelink_fromnode,
                       np.array([1, 2, 3, 6, 7, 8, 11, 12, 13,
                                 5, 6, 7, 8, 10, 11, 12, 13]))


@with_setup(setup_grid)
def test_active_link_tonode():
    assert_array_equal(rmg.activelink_tonode,
                       np.array([6, 7, 8, 11, 12, 13, 16, 17, 18,
                                 6, 7, 8, 9, 11, 12, 13, 14]))


@with_setup(setup_grid)
def test_active_link_num_inlink():
    assert_array_equal(rmg.node_numactiveinlink,
                       np.array([0, 0, 0, 0, 0,
                                 0, 2, 2, 2, 1,
                                 0, 2, 2, 2, 1,
                                 0, 1, 1, 1, 0]))


@with_setup(setup_grid)
def test_active_link_num_outlink():
    assert_array_equal(rmg.node_numactiveoutlink, np.array([0, 1, 1, 1, 0,
                                                            1, 2, 2, 2, 0,
                                                            1, 2, 2, 2, 0,
                                                            0, 0, 0, 0, 0]))


@with_setup(setup_grid)
def test_active_inlink_matrix():
    assert_array_equal(rmg.node_active_inlink_matrix,
                       np.array([[-1, -1, -1, -1, -1,
                                  -1,  0,  1,  2, -1,
                                  -1,  3,  4,  5, -1,
                                  -1,  6, 7,  8, -1],
                                 [-1, -1, -1, -1, -1,
                                  -1,  9, 10, 11, 12,
                                  -1, 13, 14, 15, 16,
                                  -1, -1, -1, -1, -1]]))


@with_setup(setup_grid)
def test_active_outlink_matrix():
    assert_array_equal(
        rmg.node_active_outlink_matrix,
        np.array([[-1,  0,  1,  2, -1,
                   -1,  3,  4,  5, -1,
                   -1,  6,  7,  8, -1,
                   -1, -1, -1, -1, -1],
                  [-1, -1, -1, -1, -1,
                    9, 10, 11, 12, -1,
                   13, 14, 15, 16, -1,
                   -1, -1, -1, -1, -1]]))


@with_setup(setup_grid)
def test_active_node_links_scalar_interior():
    assert_array_equal(rmg.active_node_links([6]), np.array([[0, 9, 3, 10]]).T)


@with_setup(setup_grid)
def test_active_node_links_scalar_boundary():
    assert_array_equal(rmg.active_node_links([1]),
                       np.array([[-1, -1, 0, -1]]).T)


@with_setup(setup_grid)
def test_active_node_with_array_arg():
    assert_array_equal(rmg.active_node_links([6, 7]),
                       np.array([[0,  9, 3, 10],
                                 [1, 10, 4, 11]]).T)


@with_setup(setup_grid)
def test_active_node_links_with_no_args():
    assert_array_equal(
        rmg.active_node_links(),
        np.array([[-1, -1, -1, -1, -1, -1,  0,  1,  2, -1,
                   -1,  3,  4,  5, -1, -1,  6,  7,  8, -1],
                  [-1, -1, -1, -1, -1, -1,  9, 10, 11, 12,
                   -1, 13, 14, 15, 16, -1, -1, -1, -1, -1],
                  [-1,  0,  1,  2, -1, -1,  3,  4,  5, -1,
                   -1,  6,  7,  8, -1, -1, -1, -1, -1, -1],
                  [-1, -1, -1, -1, -1,  9, 10, 11, 12, -1,
                   13, 14, 15, 16, -1, -1, -1, -1, -1, -1]]))


@with_setup(setup_grid)
def test_link_fromnode():
    assert_array_equal(
        rmg.link_fromnode,
        np.array([0, 1, 2, 3, 4, 5, 6, 7,  8,  9, 10, 11, 12, 13, 14,
                  0, 1, 2, 3, 5, 6, 7, 8, 10, 11, 12, 13, 15, 16, 17, 18]))


@with_setup(setup_grid)
def test_link_tonode():
    assert_array_equal(
        rmg.link_tonode,
        np.array([5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
                  1, 2, 3, 4, 6,  7,  8,  9, 11, 12, 13, 14, 16, 17, 18, 19]))


@with_setup(setup_grid)
def test_link_num_inlink():
    assert_array_equal(rmg.node_numinlink,
                       np.array([0, 1, 1, 1, 1,
                                 1, 2, 2, 2, 2,
                                 1, 2, 2, 2, 2,
                                 1, 2, 2, 2, 2]))


@with_setup(setup_grid)
def test_link_num_outlink():
    assert_array_equal(rmg.node_numoutlink, np.array([2, 2, 2, 2, 1,
                                                      2, 2, 2, 2, 1,
                                                      2, 2, 2, 2, 1,
                                                      1, 1, 1, 1, 0]))


@with_setup(setup_grid)
def test_node_inlink_matrix():
    assert_array_equal(rmg.node_inlink_matrix,
                       np.array([[-1, -1, -1, -1, -1,
                                   0,  1,  2,  3,  4,
                                   5,  6,  7,  8,  9,
                                  10, 11, 12, 13, 14],
                                 [-1, 15, 16, 17, 18,
                                  -1, 19, 20, 21, 22,
                                  -1, 23, 24, 25, 26,
                                  -1, 27, 28, 29, 30]]))


@with_setup(setup_grid)
def test_node_outlink_matrix():
    assert_array_equal(rmg.node_outlink_matrix,
                       np.array([[ 0,  1,  2,  3,  4,
                                  5,  6,  7,  8,  9,
                                  10, 11, 12, 13, 14,
                                  -1, -1, -1, -1, -1],
                                 [15, 16, 17, 18, -1,
                                  19, 20, 21, 22, -1,
                                  23, 24, 25, 26, -1,
                                  27, 28, 29, 30, -1]]))


@with_setup(setup_grid)
def test_node_links_with_scalar_interior():
    assert_array_equal(rmg.node_links([6]),
                       np.array([[1, 19, 6, 20]]).T)


@with_setup(setup_grid)
def test_node_links_with_scalar_boundary():
    assert_array_equal(rmg.node_links([1]), np.array([[-1, 15, 1, 16]]).T)


@with_setup(setup_grid)
def test_node_links_with_array_arg():
    assert_array_equal(rmg.node_links([6, 7]),
                       np.array([[1, 19, 6, 20], [2, 20, 7, 21]]).T)


@with_setup(setup_grid)
def test_node_links_with_no_args():
    assert_array_equal(
        rmg.node_links(),
        np.array([[-1, -1, -1, -1, -1,  0,  1,  2,  3,  4,
                    5,  6,  7,  8,  9, 10, 11, 12, 13, 14],
                  [-1, 15, 16, 17, 18, -1, 19, 20, 21, 22,
                   -1, 23, 24, 25, 26, -1, 27, 28, 29, 30],
                  [ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9,
                   10, 11, 12, 13, 14, -1, -1, -1, -1, -1],
                  [15, 16, 17, 18, -1, 19, 20, 21, 22, -1,
                   23, 24, 25, 26, -1, 27, 28, 29, 30, -1]]))


@with_setup(setup_grid)
def test_link_face():
    assert_array_equal(rmg.link_face, np.array([X, 0, 1, 2, X,
                                                X, 3, 4, 5, X,
                                                X, 6, 7, 8, X,
                                                X, X, X, X,
                                                9, 10, 11, 12,
                                                13, 14, 15, 16,
                                                X, X, X, X]))


@with_setup(setup_grid)
def test_grid_coords_to_node_id_with_scalar():
    assert_equal(rmg.grid_coords_to_node_id(3, 4), 19)


@with_setup(setup_grid)
def test_grid_coords_to_node_id_with_array():
    assert_array_equal(rmg.grid_coords_to_node_id((3, 2), (4, 1)),
                       np.array([19, 11]))


@with_setup(setup_grid)
def test_grid_coords_to_node_id_outside_of_grid():
    assert_raises(ValueError, rmg.grid_coords_to_node_id, 5, 0)


@with_setup(setup_grid)
def test_create_diagonal_list():
    rmg.create_diagonal_list()

    assert_array_equal(
        rmg.get_diagonal_list(),
        np.array([[6, X, X, X], [7, 5, X, X], [8, 6, X, X],
                    [9, 7, X, X], [X, 8, X, X],
                  [11, X, X, 1], [12, 10,  0,  2], [13, 11,  1,  3],
                    [14, 12,  2,  4], [X, 13, 3, X],
                  [16, X, X, 6], [17, 15,  5,  7], [18, 16,  6,  8],
                    [19, 17,  7,  9], [X, 18, 8, X],
                  [X, X, X, 11], [X, X, 10, 12], [X, X, 11, 13],
                    [X, X, 12, 14], [X, X, 13, X]]))



import numpy as np
from numpy.testing import assert_array_equal
from nose.tools import with_setup

from landlab import RasterModelGrid
from landlab.grid.base import BAD_INDEX_VALUE


def test_all_neighbors():
    rmg = RasterModelGrid(5, 4)
    X = BAD_INDEX_VALUE
    expected = np.array([
        [X, X, X, X], [X, 5, X, X], [X, 6, X, X], [X, X, X, X],
        [5, X, X, X], [6, 9, 4, 1], [7,10, 5, 2], [X, X, 6, X],
        [9, X, X, X], [10,13, 8, 5], [11,14, 9, 6], [X, X, 10, X],
        [13, X, X, X], [14,17,12, 9], [15,18,13,10], [X, X,14, X],
        [X, X, X, X], [X, X, X,13], [X, X, X,14], [X, X, X, X],
    ])
    assert_array_equal(rmg.get_neighbor_list(), expected)


def test_neighbor_list_with_scalar_arg():
    X = BAD_INDEX_VALUE
    rmg = RasterModelGrid(5, 4)

    assert_array_equal(rmg.get_neighbor_list(6), np.array([7, 10, 5, 2]))
    assert_array_equal(rmg.get_neighbor_list(-1), np.array([X, X, X, X]))
    assert_array_equal(rmg.get_neighbor_list(-2), np.array([X, X, X, 14]))


def test_neighbor_list_with_array_arg():
    X = BAD_INDEX_VALUE
    rmg = RasterModelGrid(5, 4)
    assert_array_equal(rmg.get_neighbor_list([6, -1]),
                       np.array([[7, 10, 5, 2], [X, X, X, X]]))


def test_neighbor_list_boundary():
    """
    All of the neighbor IDs for a boundary cell are -1.
    """
    X = BAD_INDEX_VALUE
    rmg = RasterModelGrid(5, 4)
    import landlab.utils.structured_grid as sgrid
    rmg.set_inactive_nodes([0, 1, 2, 3, 4, 7, 8, 11, 12, 15, 16, 17, 18, 19])

    for node_id in sgrid.boundary_iter(rmg.shape):
        assert_array_equal(rmg.get_neighbor_list(node_id),
                           np.array([X, X, X, X]))



#! /usr/env/python

"""
route_flow_dn.py:
    
Given a ModelGrid, calculates single-path (steepest direction) flow directions,
drainage area, and (optionally) discharge.

The "dn" in the name means that this is a generalization of the D8 algorithm,
for a grid in which a node has N neighbors (N might happen to be 8, or not).

Created GT Nov 2013
Modified to save data to grid directly, DEJH March 2014
"""

import landlab
#from landlab import RasterModelGrid
from landlab.components.flow_routing import flow_direction_DN
#reload(flow_direction_DN)
from landlab.components.flow_accum import flow_accum_bw
from landlab import FieldError, Component
from landlab import ModelParameterDictionary
import numpy
#from scipy.weave.build_tools import CompileError

#output_suppression_flag = True

class FlowRouter(Component):
    """
    This class implements single-path (steepest direction) flow routing, and 
    calculates flow directions, drainage area, and (optionally) discharge. 
    
    It initializes with a reference to a ModelGrid of any kind. Optionally, it
    can also take *input_params*, the string which is the name of a text input
    file. The input file can optionally contain 'runoff_rate', a float giving
    a (spatially constant) runoff rate. This is equivalent to the optional 
    input field 'water__volume_flux_in', and will override it if both are set.
    If neither are set, value will default to 1.
    
    The primary method of this class is :func:`route_flow`.
    """
    
    _name = 'DNFlowRouter'
    
    _input_var_names = set(['topographic_elevation',
                            'water__volume_flux_in',
                            ])
    
    _output_var_names = set(['drainage_area',
                             'flow_receiver',
                             'topographic__steepest_slope',
                             'water__volume_flux',
                             'upstream_ID_order',
                             'links_to_flow_receiver',
                             'flow_sinks',
                             ])
                             
    _var_units = {'topographic_elevation' : 'm',
                  'water__volume_flux_in' : 'm**3/s',
                  'drainage_area' : 'm**2',
                  'flow_receiver' : '-',
                  'topographic__steepest_slope' : '-',
                  'water__volume_flux' : 'm**3/s',
                  'upstream_ID_order' : '-',
                  'links_to_flow_receiver' : '-',
                  'flow_sinks' : '-',
                  }
    
    _var_mapping = {'topographic_elevation' : 'node',
                    'water__volume_flux_in' : 'node',
                    'drainage_area' : 'node',
                    'flow_receiver' : 'node',
                    'topographic__steepest_slope' : 'node',
                    'water__volume_flux' : 'node',
                    'upstream_ID_order' : 'node',
                    'links_to_flow_receiver' : 'node',
                    'flow_sinks' : 'node',
                    }
    
    _var_defs = {'topographic_elevation' : 'Land surface topographic elevation',
                 'water__volume_flux_in' : 'External volume water input to each node (e.g., rainfall)',
                 "drainage_area" : "Upstream accumulated surface area contributing to the node's discharge",
                 'flow_receiver' : 'Node array of receivers (node that receives flow from current node)',
                 'topographic__steepest_slope' : 'Node array of steepest *downhill* slopes',
                 'water__volume_flux' : 'Discharge of water through each node',
                 'upstream_ID_order' : 'Node array containing downstream-to-upstream ordered list of node IDs',
                 'links_to_flow_receiver' : 'ID of link downstream of each node, which carries the discharge',
                 'flow_sinks' : 'Boolean array, True at local lows',
                  }
    
    
    def __init__(self, model_grid, input_params=None):
        
        # We keep a local reference to the grid
        self._grid = model_grid
        self.value_field = 'topographic_elevation'
        
        if input_params:
            if type(input_params) == str:
                input_dict = ModelParameterDictionary(input_params)
            else:
                assert type(input_params) == dict
                input_dict = input_params
        
        # We'll also keep track of the active links; if raster, then these are
        # the "D8" links; otherwise, it's just activelinks
        if type(model_grid) is landlab.grid.raster.RasterModelGrid:
            dal, d8f, d8t = model_grid.d8_active_links()
            self._active_links = dal
            self._activelink_from = d8f
            self._activelink_to = d8t
        else:
            self._active_links = model_grid.active_links
            self._activelink_from = model_grid.activelink_fromnode
            self._activelink_to = model_grid.activelink_tonode
        
        #test input variables are present:
        model_grid.at_node['topographic_elevation']
        try:
            model_grid.at_node['water__volume_flux_in']
            self.field_for_runoff = True
        except FieldError:
            self.field_for_runoff = False
            #build the input array into the grid. This is important in case variable values appear during model run
            model_grid.add_ones('node', 'water__volume_flux_in')
            
        if not self.field_for_runoff:
            try:
                model_grid.at_node['water__volume_flux_in'].fill(input_dict['runoff_rate'])
            except (KeyError, UnboundLocalError):
                model_grid.at_node['water__volume_flux_in'].fill(1.)
        else:
            try:
                model_grid.at_node['water__volume_flux_in'].fill(input_dict['runoff_rate'])
            except (KeyError, UnboundLocalError):
                pass
            else:
                print "WARNING: Both a field and input parameter are available for runoff value. Was this intentional?? Taking the input parameter value..."
        
        # Keep track of the following variables:
        #   - drainage area at each node
        #   - receiver of each node
        self.drainage_area = model_grid.add_zeros('node', 'drainage_area')
        self.receiver = model_grid.create_node_array_zeros('flow_receiver')
        self.steepest_slope = model_grid.create_node_array_zeros('topographic__steepest_slope')
        self.discharges = model_grid.create_node_array_zeros('water__volume_flux')
        self.upstream_ordered_nodes = model_grid.create_node_array_zeros('upstream_ID_order')
        self.links_to_receiver = model_grid.create_node_array_zeros('links_to_flow_receiver')
        
        self.weave_flag = model_grid.weave_flag
        
        
    def route_flow(self):
        """
        Routes surface-water flow by (1) assigning to each node a single 
        drainage direction, and then (2) adding up the number of nodes that
        contribute flow to each node on the grid (including the node itself).
                
        Stores as ModelGrid fields:
            - Node array of receivers (nodes that receive flow): 
              *'flow_receiver'*
            - Node array of drainage areas: *'drainage_area'*
            - Node array of discharges: *'water__volume_flux'*
            - Node array of steepest downhill slopes: *'topographic__steepest_slope'*
            - Node array containing downstream-to-upstream ordered list of node
              IDs: *'upstream_ID_order'*
            - Node array containing ID of link that leads from each node to its
              receiver (or ITS OWN ID if there is no receiver):
              *'links_to_flow_receiver'*
            - Boolean node array of all local lows: *'flow_sinks'*
        
        Returns:
            - the modified grid object
        
        Examples
        --------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid
        >>> from landlab.components.flow_routing.route_flow_dn import FlowRouter
        >>> mg = RasterModelGrid(5, 4, 1.0)
        >>> elev = np.array([0.,  0.,  0., 0.,
        ...                  0., 21., 10., 0.,
        ...                  0., 31., 20., 0.,
        ...                  0., 32., 30., 0.,
        ...                  0.,  0.,  0., 0.])
        >>> _ = mg.add_field('node','topographic_elevation', elev)
        >>> mg.set_closed_boundaries_at_grid_edges(False, True, True, True)
        >>> fr = FlowRouter(mg)
        >>> mg = fr.route_flow()
        >>> mg.at_node['flow_receiver']
        array([ 0,  1,  2,  3,  4,  1,  2,  7,  8,  6,  6, 11, 12, 10, 10, 15, 16,
               17, 18, 19])
        >>> mg.at_node['drainage_area']
        array([ 1.,  2.,  6.,  1.,  1.,  1.,  5.,  1.,  1.,  1.,  3.,  1.,  1.,
                1.,  1.,  1.,  1.,  1.,  1.,  1.])

        Now let's change the cell area and the runoff rates:
        
        >>> mg = RasterModelGrid(5, 4, 10.) #so cell area==100.
        >>> _ = mg.add_field('node','topographic_elevation', elev) #put the data back into the new grid
        >>> mg.set_closed_boundaries_at_grid_edges(False, True, True, True)
        >>> fr = FlowRouter(mg)
        >>> runoff_rate = np.arange(mg.number_of_nodes)
        >>> _ = mg.add_field('node', 'water__volume_flux_in', runoff_rate)
        >>> mg = fr.route_flow()
        >>> mg.at_node['water__volume_flux']
        array([    0.,   600.,  5400.,   300.,   400.,   500.,  5200.,   700.,
                 800.,   900.,  3700.,  1100.,  1200.,  1300.,  1400.,  1500.,
                1600.,  1700.,  1800.,  1900.])
        
        """
        
        #if elevs is not provided, default to stored grid values, which must be provided as grid
        elevs = self._grid['node'][self.value_field]
        
        node_cell_area = self._grid.forced_cell_areas
            
        
        # Calculate the downhill-positive slopes at the d8 active links
        #TODO: generalize to use EITHER D8, if raster, or just active links,
        # otherwise.
        link_slope = -self._grid.calculate_gradients_at_d8_active_links(elevs)
        # Find the baselevel nodes
        (baselevel_nodes, ) = numpy.where(numpy.logical_or(self._grid.node_status==1, self._grid.node_status==2))

        # Calculate flow directions
        receiver, steepest_slope, sink, recvr_link  = \
            flow_direction_DN.flow_directions(elevs, self._active_links, 
                                         self._activelink_from,
                                         self._activelink_to, link_slope, 
                                         grid=self._grid,
                                         baselevel_nodes=baselevel_nodes, 
                                         use_weave=self.weave_flag)
#############grid=None???
        
        # TODO: either need a way to calculate and return the *length* of the
        # flow links, OR the caller has to handle the raster / non-raster case.
        
        #print 'sinks:', sink

        # Calculate drainage area, discharge, and ...
        a, q, s = flow_accum_bw.flow_accumulation(receiver, sink,
                                                  node_cell_area=node_cell_area, 
                                                  runoff_rate=self._grid.at_node['water__volume_flux_in'],
                                                  use_weave=self.weave_flag)
                                                  
        #added DEJH March 2014:
        #store the generated data in the grid
        self._grid['node']['drainage_area'] = a
        self._grid['node']['flow_receiver'] = receiver
        self._grid['node']['topographic__steepest_slope'] = steepest_slope
        self._grid['node']['water__volume_flux'] = q
        self._grid['node']['upstream_ID_order'] = s
        self._grid['node']['links_to_flow_receiver'] = recvr_link
        self._grid['node']['flow_sinks'] = numpy.zeros_like(receiver, dtype=bool)
        self._grid['node']['flow_sinks'][sink] = True
        
        return self._grid

    @property
    def node_drainage_area(self):
        return self._grid['node']['drainage_area']

    @property
    def node_receiving_flow(self):
        return self._grid['node']['flow_receiver']

    @property
    def node_steepest_slope(self):
        return self._grid['node']['topographic__steepest_slope']

    @property
    def node_water_discharge(self):
        return self._grid['node']['water__volume_flux']

    @property
    def node_order_upstream(self):
        return self._grid['node']['upstream_ID_order']

    @property
    def link_to_flow_receiving_node(self):
        return self._grid['node']['links_to_flow_receiver']


if __name__ == '__main__':
    import doctest
    doctest.testmod()



#! /usr/bin/python

from landlab import Component
import numpy as np
import datetime as datetime
from scipy.sparse import csr_matrix
from scipy.sparse import linalg


class Glacier(Component):

	def __init__(self,grid,dictionary,**kwds):
		'''
		Define physical parameters (here assuming EISMINT-1 values)
		'''
		self.n_GLEN = 3          # Glen's flow law exponent
		self.A_GLEN = 7.5738e-17 #6.05904e-18; Monthly #7.5738e-17 Cuffey & Paterson (4th ed) Glen's law parameter in Pa^{-3} yr^{-1} units (same as A_GLEN=2.4e-24 Pa^{-3} s^{-1})

		self.m_SLIDE = 2        # Sliding law exponent
		self.C_SLIDE = 0    # 1.0e-08;  # 1.0e-06;  # Sliding coefficient in Pa, metre,(Year units)

		self.RHO = 900   # Density (SI units)
		self.g = 9.80    # Gravity (SI units, rho*g has units of Pa)
		self.K_eps = 1.0e-12
		self.OMEGA = 1.5  # 1.6
		super(Glacier,self).__init__(grid)
		self.initialize(dictionary)

	def initialize(self,dictionary,**kwds):
		'''
		Initialize values for calculation:

		S: ice surface ice_elevation
		B: bed elevation 
		b_dot: mass of ice added or subtracted from each cell 
		dt: time interval
		t_STOP: ending time for modeling
		t: starting time for modeling
		dx: node spacing
		nx: number of columns of nodes
		ny: number of rows of nodes
		N: number of nodes
		'''
		self.S = kwds.pop('S',dictionary['S'])
		self.B = kwds.pop('B',dictionary['B'])
		self.b_dot = kwds.pop('b_dot',dictionary['b_dot'])
		self.dt = kwds.pop('dt',dictionary['dt'])
		self.t_STOP = kwds.pop('t_STOP',dictionary['t_STOP'])
		self.t = kwds.pop('t',dictionary['t'])
		self.dx = kwds.pop('dx',dictionary['dx'])
		self.nx = kwds.pop('nx',dictionary['nx'])
		self.ny = kwds.pop('ny',dictionary['ny'])
		self.N = self.nx * self.ny 
		self.setupIndexArrays()

	def step_update(self):
		'''
		calculate S (ice surface elevation) for each timestep 

		H_max: maximum ice thickness
		S_max: maximum ice surface elevation
		'''
		self.S, self.t = self.step()
		SB = self.S - self.B
		self.H_max = np.max(SB)
		self.k_H_max = np.argmax(SB)
		self.S_max = np.max(self.S)
		self.k_S_max = np.argmax(self.S)		

	def recursive_steps(self):
		'''
		Iterate over each time step to update the ice surface elevation 
		'''
		while 1:
			self.step_update()
			# print self.S[0:5]
			self.ALPHA_I = 100*np.sum(self.S > self.B)/float(self.N)
			
			print 'BKS: At t={:8.2f} yr ALPHA_I={:.2f}% and maxima are: H({:d}) = {:f} \
			S({:d})={:f}\n'.format(self.t, self.ALPHA_I, self.k_H_max, self.H_max, self.k_S_max, self.S_max)

			### Stop iterating until the final timestep
			if self.t > self.t_STOP:
				I = np.zeros(self.N)
				I[self.S > self.B] = 1

				# S_map = self.S.reshape(self.ny,self.nx)
				# B_map = self.B.reshape(self.ny,self.nx)
				# I_map = I.reshape(self.ny,self.nx) 

				### Note: the difference between python and matlab in matrix orders
				S_map = self.S.reshape(self.nx,self.ny).T
				B_map = self.B.reshape(self.nx,self.ny).T
				I_map = I.reshape(self.nx,self.ny).T
				H_map = S_map - B_map
				self.grid['node']['ice_elevation'] = S_map
				self.grid['node']['B_map'] = B_map
				self.grid['node']['I_map'] = I_map
				self.grid['node']['ice_thickness'] = H_map
				now = datetime.datetime.now().strftime('%H:%M:%S')
				file_str = 'S_map.txt'
				print 'main(): Output stored in file "{:s}" at time {:s} \n'.format(file_str,now)
				break

	def step(self): 
		'''
		For each timestep, a sparse linear system (Ax = C) need to be solved to update ice surface elevation
		'''

		### update diffusivity for each timestep
		self.diffusion_update()
		D_sum = self.D_IC_jc + self.D_IP_jc + self.D_ic_JC + self.D_ic_JP
		
		row = np.int64([[self.ic_jc],[self.ic_jc],[self.ic_jc],[self.ic_jc],[self.ic_jc]]).flatten()
		col = np.int64([[self.im_jc],[self.ip_jc],[self.ic_jm],[self.ic_jp],[self.ic_jc]]).flatten()
		val = np.array([[-self.OMEGA * self.D_IC_jc],[-self.OMEGA * self.D_IP_jc],[-self.OMEGA * self.D_ic_JC],[-self.OMEGA * self.D_ic_JP],[1/self.dt + self.OMEGA * D_sum]]).flatten()
		C = (1 - self.OMEGA) * ((self.D_IC_jc * self.S[self.im_jc]) + self.D_IP_jc * self.S[self.ip_jc] + self.D_ic_JC * self.S[self.ic_jm] + self.D_ic_JP * \
			self.S[self.ic_jp]) + (1/self.dt - (1 - self.OMEGA) * D_sum) * self.S[self.ic_jc] + self.b_dot 
		C = C.flatten()	

		### construct a sparse matrix A
		A = csr_matrix( (val,(row,col)), shape=(self.N, self.N))
		# print 'solving'
		S_out = linalg.spsolve(A,C)
		# print 'solved'

		### ice thickness couldn't be negative, ice surface elevation should not be less than bed elevation
		S_out[S_out < self.B] = self.B[S_out < self.B]	

		t_n = self.t + self.dt
		return S_out, t_n

	def diffusion_update(self):
		'''
		calculate diffusivity for each timestep
		'''
		A_tilde = 2 * self.A_GLEN * (self.RHO * self.g) ** self.n_GLEN/(self.n_GLEN + 2)/(self.dx ** 2)
		C_tilde = self.C_SLIDE * (self.RHO * self.g)**self.m_SLIDE/(self.dx**2)
		nm_half = (self.n_GLEN - 1) / 2.0   ### @
		npl = self.n_GLEN + 1
		mm_half = (self.m_SLIDE - 1) / 2.0  ### @
		ml = self.m_SLIDE
		
		SB = self.S - self.B
		SB[SB<0] = 0
		H = SB
		
		H_IC_jc = 0.5*(H[self.ic_jc] + H[self.im_jc])
		H_ic_JC = 0.5*(H[self.ic_jc] + H[self.ic_jm])
		
		H_IC_jc_up = H[self.im_jc]
		H_ic_JC_up = H[self.ic_jm]
		
		ix = (self.S[self.ic_jc]>self.S[self.im_jc]).reshape(-1)
		H_IC_jc_up[self.S[self.ic_jc]>self.S[self.im_jc]] = H[self.ic_jc[ix]].reshape(-1)

		ix = (self.S[self.ic_jc]>self.S[self.ic_jm]).reshape(-1)
		H_ic_JC_up[self.S[self.ic_jc]>self.S[self.ic_jm]] = H[self.ic_jc[ix]].reshape(-1)
		
		dS_dx_IC_jc = (self.S[self.ic_jc] - self.S[self.im_jc])/self.dx
		dS_dy_IC_jc = (self.S[self.ic_jp] + self.S[self.im_jp] - self.S[self.ic_jm] - self.S[self.im_jm])/(4*self.dx)
		dS_dx_ic_JC = (self.S[self.ip_jc] + self.S[self.ip_jm] - self.S[self.im_jc] - self.S[self.im_jm])/(4*self.dx)
		dS_dy_ic_JC = (self.S[self.ic_jc] - self.S[self.ic_jm])/self.dx
		
		S2_IC_jc = np.square(dS_dx_IC_jc) + np.square(dS_dy_IC_jc) + self.K_eps
		S2_ic_JC = np.square(dS_dx_ic_JC) + np.square(dS_dy_ic_JC) + self.K_eps
		
		if C_tilde == 0:    ### No sliding case
			self.D_IC_jc = A_tilde*H_IC_jc_up*np.power(H_IC_jc,npl)*np.power(S2_IC_jc,nm_half)
			self.D_ic_JC = A_tilde*H_ic_JC_up*np.power(H_ic_JC,npl)*np.power(S2_ic_JC,nm_half)
		elif C_tilde > 0:    ### Sliding case
			self.D_IC_jc = A_tilde*H_IC_jc_up*np.power(H_IC_jc,npl)*np.power(S2_IC_jc,nm_half) \
					+ C_tilde*H_IC_jc_up*np.power(H_IC_jc,ml)*np.power(S2_IC_jc,mm_half)
			self.D_ic_JC = A_tilde*H_ic_JC_up*np.power(H_ic_JC,npl)*np.power(S2_ic_JC,nm_half) \
					+ C_tilde*H_ic_JC_up*np.power(H_ic_JC,ml)*np.power(S2_ic_JC,mm_half)
		else:
			print 'diffusion(): C_tilde is undefined or incorrectly defined'
			
		self.D_IP_jc  = self.D_IC_jc[self.ip_jc]
		self.D_ic_JP  = self.D_ic_JC[self.ic_jp]
		
	def setupIndexArrays(self):
		
		ic = np.arange(self.nx)
		ip = np.append(np.array([np.arange(1,self.nx)]),self.nx - 1)
		im = np.append(0,np.array([np.arange(self.nx - 1)]))

		jc = np.arange(self.ny)
		jp = np.append(0,np.array([np.arange(self.ny - 1)]))
		jm = np.append(np.array([np.arange(1,self.ny)]),self.ny - 1)

		ic_jc = np.arange(1, self.N + 1).reshape(self.nx, self.ny)

		self.ip_jc = self.setupArrays(ip,jc,ic_jc) - 1
		self.im_jc = self.setupArrays(im,jc,ic_jc) - 1
		self.ic_jp = self.setupArrays(ic,jp,ic_jc) - 1
		self.ic_jm = self.setupArrays(ic,jm,ic_jc) - 1

		self.im_jm = self.setupArrays(im,jm,ic_jc) - 1
		self.ip_jm = self.setupArrays(ip,jm,ic_jc) - 1
		self.im_jp = self.setupArrays(im,jp,ic_jc) - 1
		self.ip_jp = self.setupArrays(ip,jp,ic_jc) - 1

		self.ic_jc = ic_jc.reshape(-1) - 1


	def setupArrays(self,a, b, ic_jc):
		x,y = np.meshgrid(b,a)
		array = []
		for l in zip(y.ravel(),x.ravel()):
			array.append(ic_jc[l])
		array = np.array(array)
		return array



#! /usr/bin/env python

from nose.tools import assert_true, assert_false, assert_raises
try:
    from nose.tools import assert_is, assert_set_equal, assert_dict_equal
except ImportError:
    from landlab.testing.tools import (assert_is, assert_set_equal,
                                       assert_dict_equal)
import numpy as np
from numpy.testing import assert_array_equal

from landlab.field.grouped import ModelDataFields, GroupError


def test_init():
    fields = ModelDataFields()
    assert_set_equal(set(), fields.groups)


def test_new_field_location():
    fields = ModelDataFields()
    fields.new_field_location('node', 12)
    assert_set_equal(set(['node']), fields.groups)


def test_add_existing_group():
    fields = ModelDataFields()
    fields.new_field_location('node', 12)
    assert_raises(ValueError, fields.new_field_location, 'node', 24)


def test_add_multiple_groups():
    fields = ModelDataFields()
    fields.new_field_location('node', 12)
    fields.new_field_location('cell', 2)
    fields.new_field_location('face', 7)
    fields.new_field_location('link', 7)
    assert_set_equal(set(['node', 'cell', 'face', 'link']), fields.groups)

def test_ones():
    fields = ModelDataFields()
    fields.new_field_location('node', 12)
    fields.new_field_location('cell', 2)

    value_array = fields.ones('node')
    assert_array_equal(np.ones(12), value_array)

    value_array = fields.ones('cell')
    assert_array_equal(np.ones(2), value_array)

def test_add_ones():
    fields = ModelDataFields()
    fields.new_field_location('node', 12)
    fields.new_field_location('cell', 2)

    fields.add_ones('node', 'z')
    assert_array_equal(np.ones(12), fields['node']['z'])
    assert_array_equal(np.ones(12), fields.field_values('node', 'z'))

    fields.add_ones('cell', 'z')
    assert_array_equal(np.ones(2), fields['cell']['z'])
    assert_array_equal(np.ones(2), fields.field_values('cell', 'z'))

def test_add_ones_return_value():
    fields = ModelDataFields()
    fields.new_field_location('node', 12)
    fields.new_field_location('cell', 2)

    rtn_value = fields.add_ones('node', 'z')
    assert_array_equal(rtn_value, np.ones(12))
    assert_is(rtn_value, fields['node']['z'])
    assert_is(rtn_value, fields.field_values('node', 'z'))

    rtn_value = fields.add_ones('cell', 'z')
    assert_array_equal(rtn_value, np.ones(2))
    assert_is(rtn_value, fields['cell']['z'])
    assert_is(rtn_value, fields.field_values('cell', 'z'))

def test_getitem():
    fields = ModelDataFields()
    fields.new_field_location('node', 12)

    assert_dict_equal(dict(), fields['node'])
    assert_raises(GroupError, lambda k: fields[k], 'cell')
    assert_raises(KeyError, lambda k: fields[k], 'cell')


def test_at_attribute():
    fields = ModelDataFields()
    fields.new_field_location('node', 12)

    assert_dict_equal(dict(), fields.at_node)
    assert_raises(AttributeError, lambda: fields.at_cell)

    fields.add_ones('node', 'z')
    assert_array_equal(np.ones(12), fields.at_node['z'])

def test_has_group():
    fields = ModelDataFields()
    fields.new_field_location('node', 12)

    assert_true(fields.has_group('node'))
    assert_false(fields.has_group('cell'))



import numpy as np
cimport numpy as np
cimport cython

from libc.math cimport exp, log


DTYPE_FLOAT = np.double
ctypedef np.double_t DTYPE_FLOAT_t

DTYPE_INT = np.int
ctypedef np.int_t DTYPE_INT_t


#cdef extern from "math.h":
#    double log(double x) nogil
#    double exp(double x) nogil


@cython.boundscheck(False)
def sai_loop_1(np.ndarray[DTYPE_FLOAT_t, ndim=1] sini,
               np.ndarray[DTYPE_FLOAT_t, ndim=1] fc,
               np.ndarray[DTYPE_FLOAT_t, ndim=1] sc,
               np.ndarray[DTYPE_FLOAT_t, ndim=1] wp,
               np.ndarray[DTYPE_FLOAT_t, ndim=1] beta,
               np.ndarray[DTYPE_FLOAT_t, ndim=1] mu,
               np.ndarray[DTYPE_FLOAT_t, ndim=1] nu,
               np.ndarray[DTYPE_FLOAT_t, ndim=1] nuw,
               np.ndarray[DTYPE_FLOAT_t, ndim=1] tfc,
               np.ndarray[DTYPE_FLOAT_t, ndim=1] tsc,
               np.ndarray[DTYPE_FLOAT_t, ndim=1] twp):
    cdef unsigned int n_nodes = sini.size
    cdef unsigned int i

    for i in range(n_nodes):
        if sini[i] >= fc[i]:
            tfc[i] = (
              (1. / (beta[i] * (mu[i] - nu[i]))) *
              (beta[i] * (fc[i] - sini[i]) +
               log((nu[i] - mu[i] +
                    mu[i] * exp(beta[i] * (sini[i] - fc[i]))) / nu[i])))
            tsc[i] = ((fc[i] - sc[i]) / nu[i]) + tfc[i]
            twp[i] = (((sc[i] - wp[i]) / (nu[i] - nuw[i])) *
                      log(nu[i] / nuw[i]) + tsc[i])



#! /usr/env/python
"""
Landlab's cellular automata modeling package.

Overview
--------

A Landlab Cellular Automaton, or LCA, implements a particular type of cellular
automaton (CA): a continuous-time stochastic CA. The approach is based on that
of Narteau et al. (2002, 2009) and Rozier and Narteau (2014). Like a normal
CA, the domain consists of a lattice of cells, each of which has a discrete
state. Unlike a conventional CA, the updating process is stochastic, and takes
place in continuous rather than discrete time. Any given pair (or "doublet")
of adjacent cell states has a certain specified probability of transition to a 
different pair of states. The transition probability is given in the form of an
average *transition rate*, :math:\lambda (with dimensions of 1/T); the actual time of 
transition is a random variable drawn from an exponential probability 
distribution with mean :math:1/\lambda

Subclasses
----------

Landlab provides for several different lattice and connection types:
    - RasterLCA: regular raster grid with transitions between horizontal and
        vertical cell pairs
    - OrientedRasterLCA: like a RasterLCA, but different transition rates can
        be assigned to vertical and horizontal pairs. This property of
        orientation can be used, for example, to implement rules representing
        gravitational attraction, or flow of a fluid with a particular direction.
    - RasterD8LCA: like a RasterLCA, but includes diagonal as well as vertical
        and horizontal cell pairs.
    - OrientedRasterD8LCA: as above but orientation also matters.
    - HexLCA: hexagonal grid
    - OrientedLCA: hexagonal grid, with transition rates allowed to vary
        according to orientation.
        
Encoding of "states"
--------------------
As in any traditional cellular automaton model, a LandlabCellularAutomaton
contains a grid of cells ("nodes" in Landlab parlance), each of which is has a
discrete state. States are represented by integers (0, 1, ... N). 

In addition, every active link has an *orientation code* and a *link state code*. 
The orientation code represents the orientation of the link in space: is it
"vertical" (aligned with the y axis), "horizontal" (aligned with x), or in some
other orientation? The number of possible orientations depends on the subclass.
The base class has only one orientation code (0) (meaning "orientation doesn't
matter), but this is overridden in some of the subclasses. For example, the
OrientedRasterLCA has two orientation codes (0 and 1, for vertical and
horizontal), while the OrientedHexLCA has three (representing the three axes
in a hex-cell / triagonal grid).

Each active link also has a *link state code*. The *state* of a link refers to
its particular combination of nodes and its orientation. For example, link 
state 1 refers to a link in which the tail-node has state 0, the head-node has
state 1, and the orientation code is 0. The number of possible link states is
equal to R N^2, where R is the number of orientations (1 to 3, depending on the 
subclass) and N is the number of possible node states. The simplest possible
Landlab CA model would have just one orientation code and two possible cell
states, so that there are four unique link states. These would be represented by
the tuples of (tail-node state, head-node state, orientation) as follows:
    
    link state 0 = (0, 0, 0)
    link state 1 = (0, 1, 0)
    link state 2 = (1, 0, 0)
    link state 3 = (1, 1, 0)
        
Main data structures
--------------------
node_state : 1d array (x number of nodes in grid)
    Node-based grid of node-state codes. This is the grid of cell (sic) states.
    
link_state_dict : dictionary
    Keys are 3-element tuples that represent the cell-state pairs and 
    orientation code for each possible link type; values are the corresponding
    link-state codes. Allows you to look up the link-state code corresponding
    to a particular pair of adjacent nodes with a particular orientation.

cell_pair : list (x number of possible link states)
    List of 3-element tuples representing all the various link states. Allows
    you to look up the node states and orientation corresponding to a particular 
    link-state ID.
    
event_queue : heap of Event objects
    Queue containing all future transition events, sorted by time of occurrence
    (from soonest to latest).
    
next_update : 1d array (x number of active links)
    Time (in the future) at which the link will undergo its next transition.
    You might notice that the update time for every scheduled transition is also
    stored in each Event object in the event queue. Why store it twice? Because
    a scheduled event might be invalidated after the event has been scheduled
    (because another transition has changed one of a link's two nodes, for
    example). The way to tell whether a scheduled event is still valid is to
    compare its time with the corresponding transition time in the *next_update*
    array. If they are different, the event is discarded.

active_link_orientation : 1d array of ints (x number of active links)
    Orientation code for each link.
    
link_state : 1d array of ints (x number of active links)
    State code for each link.

n_xn : 1d array of ints (x number of possible link states)
    Number of transitions ("xn" stands for "transition") from a given link
    state.

xn_to : 2d array of ints (# possible link states x max. # transitions)
    Stores the link-state code(s) to which a particular link state can 
    transition. "max. # transitions" means the maximum number of transitions
    from a single state. For example, if each link state is associated with one
    and only one transition, then the maximum is 1, but if there is at least
    one link state that can have either of two different transitions, then the
    maximum would be two.

xn_rate : 2d array of floats (# possible link states x max. # transitions)
    Rate associated with each link-state transition.


Created GT Sep 2014, starting from link_ca.py.
"""

from heapq import heappush
from heapq import heappop
from landlab import RasterModelGrid
import landlab
import numpy
import pylab as plt
import time

_NEVER = 1e50

_DEBUG = False

_TEST = False


class Transition():
    """
    Represents a transition from one state ("from_state") to another 
    ("to_state") at a link. The transition probability is represented by a rate 
    parameter "rate", with dimensions of 1/T. The probability distribution of 
    time until the transition event occurs is exponentional with mean 1/rate. 
    The optional name parameter allows the caller to assign a name to any given 
    transition.
    
    Note that from_state and to_state can now be either integer IDs for the
    standardised ordering of the link states (as before), or tuples explicitly
    describing the node state at each end, and the orientation.
    Orientation is 0: horizontal, L-R; 1: vertical, bottom-top.
    For such a tuple, order is (left/bottom, right/top, orientation).
    """
    def __init__(self, from_state, to_state, rate, name=None, 
                 swap_properties=False):
        """
        Transition() constructor sets 3 required properties and 2 optional
        properties for a transition from one cell pair to another.
        
        Parameters
        ----------
        from_state : int
            Code for the starting state of the cell pair (link)
        to_state : int
            Code for the new state of the cell pair (link)
        rate : float
            Average rate at which this transition occurs (dimension of 1/time)
        name : string (optional)
            Name for this transition
        swap_properties : bool (optional)
            Flag: should properties be exchanged between the two cells?
        """
        self.from_state = from_state
        self.to_state = to_state
        self.rate = rate
        self.name = name
        self.swap_properties = swap_properties
        
        
class Event():
    """
    Represents a transition event at a link. The transition occurs at a given
    link and a given time, and it involves a transition into the state xn_to
    (an integer code representing the new link state; "xn" is shorthand for
    "transition").
    
    The class overrides the __lt__ (less than operator) method so that when
    Event() objects are placed in a PriorityQueue, the earliest event is
    given the highest priority (i.e., placed at the top of the queue).
    
    Examples
    --------
    >>> from landlab.components.cellular_automata.landlab_ca import Event
    >>> e1 = Event( 10.0, 1, 2)
    >>> e2 = Event( 2.0, 3, 1)
    >>> e1 < e2
    False
    >>> e2 < e1
    True
    """
    def __init__(self, time, link, xn_to, propswap=False):
        """
        Event() constructor sets 3 required properties and one optional
        property.
        
        Parameters
        ----------
        time : float
            Time at which the event is scheduled to occur
        link : int
            ID of the link at which event occurs
        xn_to : int
            New state to which this cell pair (link) will transition
        propswap : bool (optional)
            Flag: does this event involve an exchange of properties between
            the two cells?
        """
        self.time = time
        self.link = link
        self.xn_to = xn_to
        self.propswap = propswap
        
    def __lt__(self, other):
        """
        Overridden less-than operator: returns true if the event on the left
        has an earlier scheduled time than the event on the right
        """
        return self.time < other.time
            
            
class CAPlotter():
    """
    A CAPlotter is an object that handles display of a CellLab-CTS grid.
    """
    def __init__(self, ca, cmap=None):
        """
        CAPlotter() constructor keeps a reference to the CA model, and
        optionally a colormap to be used with plots.
        
        Parameters
        ----------
        ca : LandlabCellularAutomaton object
            Reference to a CA model
        cmap : Matplotlib colormap (optional)
            Colormap to be used in plotting
        """
        import matplotlib
        
        # Set the colormap; default to matplotlib's "jet" colormap
        if cmap is None:
            self._cmap = matplotlib.cm.jet
        else:
            self._cmap = cmap
            
        # Keep a reference to the CA model
        self.ca = ca
        
        # Initialize the plot and remember the grid type
        plt.ion()
        plt.figure(1)
        if type(ca.grid) is landlab.grid.hex.HexModelGrid:
            self.gridtype = 'hex'
        else:
            self.gridtype = 'rast'


    def update_plot(self):
        """
        Plots the current node state grid.
        """
        plt.clf()
        if self.gridtype=='rast':
            nsr = self.ca.grid.node_vector_to_raster(self.ca.node_state)
            plt.imshow(nsr, interpolation='None', origin='lower', cmap=self._cmap)
        else:
            self.ca.grid.hexplot(self.ca.node_state, color_map=self._cmap)
        
        plt.draw()
        plt.pause(0.001)
        
    def finalize(self):
        """
        Wraps up plotting by switching off interactive model and showing the
        plot.
        """        
        plt.ioff()
        plt.show()
        
        
class LandlabCellularAutomaton(object):
    """
    A LandlabCellularAutomaton implements a link-type (or doublet-type) cellular
    automaton model. A link connects a pair of cells. Each cell has a state
    (represented by an integer code), and each link also has a state that is
    determined by the states of the cell pair.
    """
    def __init__(self, model_grid, node_state_dict, transition_list,
                 initial_node_states, prop_data=None, prop_reset_value=None):
        """
        LandlabCellularAutomaton() constructor initializes the CA model.
        
        Parameters
        ----------
        model_grid : Landlab ModelGrid object
            Reference to the model's grid
        node_state_dict : dict
            Keys are node-state codes, values are the names associated with
            these codes
        transition_list : list of Transition objects
            List of all possible transitions in the model
        initial_node_states : array of ints (x number of nodes in grid)
            Starting values for node-state grid
        prop_data : array (x number of nodes in grid) (optional)
            Array of properties associated with each node/cell
        prop_reset_value : (scalar; same type as entries in prop_data) (optional)
            Default or initial value for a node/cell property (e.g., 0.0)
        """
        # Are we calling this from a subclass __init__? If so, then the 
        # variable self.number_of_orientations should already be defined.
        try:
            self.number_of_orientations == 1
        except AttributeError:  # if self.number_of_orientations not already defined
            self.number_of_orientations = 1
        
        # Keep a copy of the model grid; remember how many active links in it
        self.grid = model_grid
        self.node_active_links = self.grid.node_activelinks()

        # Set up the initial node-state grid
        self.set_node_state_grid(initial_node_states)
        
        # Current simulation time starts out at zero
        self.current_time = 0.0
        
        # Figure out how many states there are, and make sure the input data
        # are self consistent.
        #   There are 2 x (N^2) link states, where N is the number of node 
        # states. For example, if there are just two node states, 0 and 1, then 
        # the possible oriented link pairs are listed below:
        #   0-0 0-1 1-0 1-1  0 0 1 1
        #                    0 1 0 1
        self.num_node_states = len(node_state_dict)
        self.num_link_states = self.number_of_orientations*self.num_node_states**2
        
        assert (type(transition_list) is list), 'transition_list must be a list!'
        assert (transition_list), \
               'Transition list must contain at least one transition'
        last_type = None
        for t in transition_list:
            try:
                assert (t.from_state < self.num_link_states), \
                    'Transition from_state out of range'
                assert (t.to_state < self.num_link_states), \
                    'Transition to_state out of range'
                this_type = int
            # TODO: make orientation optional for cases where self.number_of_orientations = 1
            except: #added to allow from and to states to be tuples, not just ids
                assert type(t.from_state) == tuple, 'Transition from_state out of range'
                assert type(t.to_state) == tuple, 'Transition to_state out of range'
                for i in t.from_state[:-1]:
                    assert (i < self.num_node_states), \
                    'Transition from_state out of range'
                for i in t.to_state[:-1]:
                    assert (i < self.num_node_states), \
                    'Transition to_state out of range'
                assert t.from_state[-1] < self.number_of_orientations, \
                    'Encoding for orientation in from_state must be < number of orientations.'
                assert t.to_state[-1] < self.number_of_orientations, \
                    'Encoding for orientation in to_state must be < number of orientations.'
                this_type = tuple
            assert last_type==this_type or last_type==None, \
                'All transition types must be either int IDs, or all tuples.'
            #this test to ensure all entries are either IDs, or tuples, not mixed
            last_type=this_type
        
        # Create priority queue for events and next_update array for links
        self.event_queue = []
        self.next_update = self.grid.create_active_link_array_zeros()
    
        # Assign link types from node types
        self.create_link_state_dict_and_pair_list()

        #DEJH adds: convert transition_list to IDs if necessary
        #This is the new part that allows Transition from_ and to_ types 
        #to be specified either as ints, or as tuples.
        transition_list_as_ID = transition_list[:]
        if type(transition_list[0].from_state) == tuple:
            #(then they all are..., because of the assertions in __init__)
            for i in range(len(transition_list)):
                transition_list_as_ID[i].from_state = self.link_state_dict[transition_list[i].from_state]
                transition_list_as_ID[i].to_state = self.link_state_dict[transition_list[i].to_state]
    
        # Set up the information needed to determine the orientation of links
        # in the lattice. The default method just creates an array of zeros (all
        # orientations considered the same), but this will be overridden
        # in subclasses that do use orientation.
        self.setup_array_of_orientation_codes()
        
        # Using the grid of node states, figure out all the link states
        self.assign_link_states_from_node_types()
    
        # Create transition data for links
        self.setup_transition_data(transition_list_as_ID)

        # Put the various transitions on the event queue
        self.push_transitions_to_event_queue()
        
        # In order to keep track of cell "properties", we create an array of
        # indices that refer to locations in the caller's code where properties
        # are tracked.
        self.propid = numpy.arange(self.grid.number_of_nodes)
        if prop_data is None:
            self.prop_data = numpy.zeros(self.grid.number_of_nodes)
            self.prop_reset_value = 0.0
        else:
            self.prop_data = prop_data
            self.prop_reset_value = prop_reset_value
        

    def set_node_state_grid(self, node_states):
        """
        Sets the grid of node-state codes to node_states. Also checks
        to make sure node_states is in the proper format, which is to
        say, it's a Numpy array of the same length as the number of nodes in 
        the grid.
        
        Parameters
        ----------
        node_states : 1D array of ints (x number of nodes in grid)
        
        Creates
        -------
        self.node_state : 1D array of ints (x number of nodes in grid)
        	The node-state array
        	
        Notes
        -----
        The node-state array is attached to the grid as a field with the name 'node_state'.
        """
        assert (type(node_states) is numpy.ndarray), \
               'initial_node_states must be a Numpy array'
        assert (len(node_states)==self.grid.number_of_nodes), \
               'length of initial_node_states must equal number of nodes in grid'
        self.grid.at_node['node_state'] = node_states
        self.node_state = node_states
        
                 
    def create_link_state_dict_and_pair_list(self):
        """
        Creates a dictionary that can be used as a lookup table to find out 
        which link state corresponds to a particular pair of node states. The 
        dictionary keys are 3-element tuples, each of which represents the state
        of the TAIL node, the HEAD node, and the orientation of the link. The 
        values are integer codes representing the link state numbers.
        """
        self.link_state_dict = {}
        self.cell_pair = []
        k=0
        for orientation in range(self.number_of_orientations):
            for tail_state in range(self.num_node_states):
                for head_state in range(self.num_node_states):
                    self.link_state_dict[(tail_state,head_state,orientation)] = k
                    k+=1
                    self.cell_pair.append((tail_state,head_state,orientation))
    
        if False and _DEBUG:
            print 
            print 'create_link_state_dict_and_pair_list(): dict is:'
            print self.link_state_dict
            print '  and the pair list is:'
            print self.cell_pair
            
            
    def setup_array_of_orientation_codes(self):
        """
        Creates and configures an array that contain the orientation code for 
        each active link (and corresponding cell pair).
        
        Parameters
        ----------
        (none)
        
        Returns
        -------
        (none)
        
        Creates
        -------
        self.active_link_orientation : 1D numpy array
        
        Notes
        -----
        The setup varies depending on the type of LCA. The default is 
        non-oriented, in which case we just have an array of zeros. Subclasses
        will override this method to handle lattices in which orientation 
        matters (for example, vertical vs. horizontal in an OrientedRasterLCA).
        """
        self.active_link_orientation = numpy.zeros(self.grid.number_of_active_links, dtype=int)
    
    
    def assign_link_states_from_node_types(self):
        """
        Assigns a link-state code for each link, and returns a list of these.
        
        Takes lists/arrays of "tail" and "head" node IDs for each link, and a 
        dictionary that associates pairs of node states (represented as a 
        3-element tuple, comprising the TAIL state, FROM state, and orientation) 
        to link states.
        """
        self.link_state = numpy.zeros(self.grid.number_of_active_links,
                                      dtype=int)
    
        for i in range(self.grid.number_of_active_links):
            orientation = self.active_link_orientation[i]
            node_pair = (self.node_state[self.grid.activelink_fromnode[i]], \
                         self.node_state[self.grid.activelink_tonode[i]], \
                         orientation)
            self.link_state[i] = self.link_state_dict[node_pair]
                    
        if False and _DEBUG:
            print 
            print 'assign_link_states_from_node_types(): the link state array is:'
            print self.link_state


    def setup_transition_data(self, xn_list):
        """
        Using the transition list and the number of link states, creates 
        three arrays that collectively contain data on state transitions:
            n_xn: for each link state, contains the number of transitions out of 
                  that state.
            xn_to: 2D array that records, for each link state and each
                   transition, the new state into which the link transitions.
            xn_rate: 2D array that records, for each link state and each
                     transition, the rate (1/time) of the transition.
            xn_propswap: 2D array that indicates, for each link state and each
                         transition, whether that transition is accompanied by
                         a "property" swap, in which the two cells exchange
                         properties (in order to represent a particle moving)
        """
        # First, create an array that stores the number of possible transitions
        # out of each state.
        self.n_xn = numpy.zeros(self.num_link_states, dtype=int)
        for xn in xn_list:
            self.n_xn[xn.from_state] += 1
        
        # Now, create arrays to hold the "to state" and transition rate for each
        # transition. These arrays are dimensioned N x M where N is the number
        # of states, and M is the maximum number of transitions from a single 
        # state (for example if state 3 could transition either to state 1 or 
        # state 4, and the other states only had one or zero possible 
        # transitions, then the maximum would be 2).
        max_transitions = numpy.max(self.n_xn)
        self.xn_to = numpy.zeros((self.num_link_states, max_transitions), dtype=int)
        self.xn_rate = numpy.zeros((self.num_link_states, max_transitions))
        self.xn_propswap = numpy.zeros((self.num_link_states, max_transitions), dtype=bool)
    
        # Populate the "to" and "rate" arrays
        self.n_xn[:] = 0  # reset this and then re-do (inefficient but should work)
        for xn in xn_list:
            from_state = xn.from_state
            self.xn_to[from_state][self.n_xn[from_state]] = xn.to_state
            self.xn_rate[from_state][self.n_xn[from_state]] = xn.rate
            self.xn_propswap[from_state][self.n_xn[from_state]] = xn.swap_properties
            self.n_xn[from_state] += 1
    
        if False and _DEBUG:
            print 
            print 'setup_transition_data():'
            print '  n_xn',self.n_xn
            print '  to:',self.xn_to
            print '  rate:',self.xn_rate
            
            
    def current_link_state(self, link_id):
        """
        Used to determines whether the link state at link *link_id* has changed 
        due to an independent change in the node-state grid. Returns the 
        current state of the link based on the states of its two end nodes; this
        can be compared to the entry in self.link_state to determine whether the
        state has changed.
        
        Parameters
        ----------
        link_id : int
            ID of the active link to test
            
        Returns
        -------
        int
            New link state code
        
        Notes
        -----
        Vectorizing this might yield some speed.
        """
        
        # Find out the states of the two nodes, and the orientation
        tail_node_state = self.node_state[self.grid.activelink_fromnode[link_id]]
        head_node_state = self.node_state[self.grid.activelink_tonode[link_id]]
        orientation = self.active_link_orientation[link_id]
        
        # Return the corresponding state code.
        return self.link_state_dict[(tail_node_state,head_node_state,orientation)]
        
        
    def update_link_states_and_transitions(self, current_time):
        """
        Following an "external" change to the node state grid, updates link
        states where necessary and creates any needed events.
        
        Algorithm
        ---------
            FOR each active link:
                if the actual node pair is different from the link's code:
                    change the link state to be correct
                    schedule an event
        """
        for i in range(self.grid.number_of_active_links):
            current_state = self.current_link_state(i)
            if current_state!=self.link_state[i]:
                self.update_link_state(i, current_state, current_time)
                
    
    def get_next_event(self, link, current_state, current_time):
        """
        Returns the next event for link with ID "link", which is in state
        "current state".
    
        Parameters
        ----------
        link : int
            ID of the link
        current_state : int
            Current state code for the link
        current_time : float
            Current time in simulation (i.e., time of event just processed)
            
        Returns
        -------
        Event object       
            The returned Event object contains the time, link ID, and type of
            the next transition event at this link.
        
        Notes
        -----        
        If there is only one potential transition out of the current state, a 
        time for the transition is selected at random from an exponential 
        distribution with rate parameter appropriate for this transition.
    
        If there are more than one potential transitions, a transition time is 
        chosen for each, and the smallest of these applied.
    
        Assumes that there is at least one potential transition from the current
        state.
        """
        assert (self.n_xn[current_state]>0), \
               'must have at least one potential transition'
    
        # Find next event time for each potential transition
        if self.n_xn[current_state]==1:
            xn = self.xn_to[current_state][0]
            propswap = self.xn_propswap[current_state][0]
            next_time = numpy.random.exponential(1.0/self.xn_rate[current_state][0])
        else:
            next_time = _NEVER
            xn = None
            propswap = False
            for i in range(self.n_xn[current_state]):
                this_next = numpy.random.exponential(1.0/self.xn_rate[current_state][i])
                if this_next < next_time:
                    next_time = this_next
                    xn = self.xn_to[current_state][i]
                    propswap = self.xn_propswap[current_state][i]

        # Create and setup event, and return it
        my_event = Event(next_time+current_time, link, xn, propswap)
    
        if _DEBUG:
            print 'get_next_event():'
            print '  next_time:',my_event.time
            print '  link:',my_event.link
            print '  xn_to:',my_event.xn_to
            print '  propswap:',my_event.propswap
    
        return my_event
    
    
    def push_transitions_to_event_queue(self):
        """
        Initializes the event queue by creating transition events for each
        cell pair that has one or more potential transitions and pushing these
        onto the queue. Also records scheduled transition times in the
        self.next_update array.
        """
        if _DEBUG:
            print 'push_transitions_to_event_queue():',self.num_link_states,self.n_xn
            
        for i in range(self.grid.number_of_active_links):
        
            if self.n_xn[self.link_state[i]] > 0:
                event = self.get_next_event(i, self.link_state[i], 0.0)
                heappush(self.event_queue, event)
                self.next_update[i] = event.time
            
            else:
                self.next_update[i] = _NEVER
            
        if _DEBUG:
            print '  push_transitions_to_event_queue(): events in queue are now:'
            for e in self.event_queue:
                print '    next_time:',e.time,'link:',e.link,'xn_to:',e.xn_to
            
            
    def update_node_states(self, tail_node, head_node, new_link_state):
        """
        Updates the states of the two nodes in the given link.
        
        Parameters
        ----------
        tail_node : int
            ID of the tail node of the link (cell pair) in question
        head_node : int
            ID of the head node of the link (cell pair) in question
        new_link_state : int
            Link state code for the new cell pair
            
        Returns
        -------
        (bool, bool)
            Flags indicating whether the tail node and head node, respectively,
            have changed state
        """
    
        # Remember the previous state of each node so we can detect whether the 
        # state has changed
        old_tail_node_state = self.node_state[tail_node]
        old_head_node_state = self.node_state[head_node]
    
        # Change to the new states
        if self.grid.node_status[tail_node]==landlab.grid.base.CORE_NODE:
            self.node_state[tail_node] = self.cell_pair[new_link_state][0]
        if self.grid.node_status[head_node]==landlab.grid.base.CORE_NODE:
            self.node_state[head_node] = self.cell_pair[new_link_state][1]
    
        if _DEBUG:
            print 'update_node_states() for',tail_node,'and',head_node
            print '  tail_node was',old_tail_node_state,'and is now',self.node_state[tail_node]
            print '  head_node was',old_head_node_state,'and is now',self.node_state[head_node]
    
        return self.node_state[tail_node]!=old_tail_node_state, \
               self.node_state[head_node]!=old_head_node_state
               

    def update_link_state(self, link, new_link_state, current_time):
        """
        Implements a link transition by updating the current state of the link
        and (if appropriate) choosing the next transition event and pushing it 
        on to the event queue.
    
        Parameters
        ----------
        link : int
            ID of the link to update
        new_link_state : int
            Code for the new state
        current_time : float
            Current time in simulation
        """
        if _DEBUG:
            print
            print 'update_link_state()'
            
        # If the link connects to a boundary, we might have a different state
        # than the one we planned
        fn = self.grid.activelink_fromnode[link]
        tn = self.grid.activelink_tonode[link]
        if _DEBUG:
            print 'fn',fn,'tn',tn,'fnstat',self.grid.node_status[fn],'tnstat',self.grid.node_status[tn]
        if self.grid.node_status[fn]!=landlab.grid.base.CORE_NODE or \
           self.grid.node_status[tn]!=landlab.grid.base.CORE_NODE:
            fns = self.node_state[self.grid.activelink_fromnode[link]]
            tns = self.node_state[self.grid.activelink_tonode[link]]
            orientation = self.active_link_orientation[link]
            actual_pair = (fns,tns,orientation)
            new_link_state = self.link_state_dict[actual_pair]
            if _DEBUG:
                print '**Boundary: overriding new link state to',new_link_state
            
        self.link_state[link] = new_link_state
        if self.n_xn[new_link_state] > 0:
            event = self.get_next_event(link, new_link_state, current_time)
            heappush(self.event_queue, event)
            self.next_update[link] = event.time
        else:
            self.next_update[link] = _NEVER
            
        if _DEBUG:
            print '  at link',link
            print '  state changed to',self.link_state[link],self.cell_pair[self.link_state[link]]
            print '  update time now',self.next_update[link]
        
            
    def do_transition(self, event, current_time, plot_each_transition=False,
                      plotter=None):
        """
        Implements a state transition. 
        
        Parameters
        ----------
        event : Event object 
        	Event object containing the data for the current transition event
        current_time : float
        	Current time in simulation
        plot_each_transition : bool (optional)
        	True if caller wants to show a plot of the grid after this transition
        plotter : CAPlotter object
        	Sent if caller wants a plot after this transition
        
        Algorithm
        ---------
        First checks that the transition is still
        valid by comparing the link's next_update time with the corresponding update
        time in the event object.
        
        If the transition is valid, we:
            1) Update the states of the two nodes attached to the link
            2) Update the link's state, choose its next transition, and push it on
            the event queue.
            3) Update the states of the other links attached to the two nodes, 
            choose their next transitions, and push them on the event queue.
            
        """

        if _DEBUG:
            print
            print 'do_transition() for link',event.link

        # We'll process the event if its update time matches the one we have 
        # recorded for the link in question. If not, it means that the link has
        # changed state since the event was pushed onto the event queue, and in that
        # case we'll ignore it.
        if event.time == self.next_update[event.link]:
        
            if _DEBUG:
                print '  event time =',event.time
            
            tail_node = self.grid.activelink_fromnode[event.link]
            head_node = self.grid.activelink_tonode[event.link]
            tail_changed, head_changed = self.update_node_states(tail_node, head_node, 
                                                          event.xn_to)
            self.update_link_state(event.link, event.xn_to, event.time)

            # Next, when the state of one of the link's nodes changes, we have to
            # update the states of the OTHER links attached to it. This could happen
            # to one or both nodes.
            if tail_changed:
                
                if _DEBUG:
                    print '    fromnode has changed state, so updating its links'
            
                for link in self.node_active_links[:,tail_node]:
                    
                    if _DEBUG:
                        print 'f checking link',link
                    if link!=-1 and link!=event.link:
                    
                        this_link_fromnode = self.grid.activelink_fromnode[link]
                        this_link_tonode = self.grid.activelink_tonode[link]
                        orientation = self.active_link_orientation[link]
                        current_pair = (self.node_state[this_link_fromnode], 
                                        self.node_state[this_link_tonode], orientation)
                        new_link_state = self.link_state_dict[current_pair]
                        self.update_link_state(link, new_link_state, event.time)

            if head_changed:
            
                if _DEBUG:
                    print '    tonode has changed state, so updating its links'
            
                for link in self.node_active_links[:,head_node]:
                
                    if _DEBUG:
                        print 't checking link',link
                    if link!=-1 and link!=event.link:
                    
                        this_link_fromnode = self.grid.activelink_fromnode[link]
                        this_link_tonode = self.grid.activelink_tonode[link]
                        orientation = self.active_link_orientation[link]
                        current_pair = (self.node_state[this_link_fromnode], 
                                        self.node_state[this_link_tonode], orientation)
                        new_link_state = self.link_state_dict[current_pair]
                        self.update_link_state(link, new_link_state, event.time)

		# If requested, display a plot of the grid
            if plot_each_transition and (plotter is not None):
                plotter.update_plot()
                
            # If this event involves an exchange of properties (i.e., the event involves
            # motion of an object that posses properties we want to track), implement
            # the swap.
            if event.propswap:
                tmp = self.propid[tail_node]
                self.propid[tail_node] = self.propid[head_node]
                self.propid[head_node] = tmp
                if self.grid.node_status[tail_node]!=landlab.grid.base.CORE_NODE:
                    self.prop_data[self.propid[tail_node]] = self.prop_reset_value
                if self.grid.node_status[head_node]!=landlab.grid.base.CORE_NODE:
                    self.prop_data[self.propid[head_node]] = self.prop_reset_value
               
            if _DEBUG:
                n = self.grid.number_of_nodes
                for r in range(self.grid.number_of_node_rows):
                    for c in range(self.grid.number_of_node_columns):
                        n -= 1
                        print '{0:.0f}'.format(self.node_state[n]),
                    print
                if self.propid is not None:
                    print
                    n = self.grid.number_of_nodes
                    for r in range(self.grid.number_of_node_rows):
                        for c in range(self.grid.number_of_node_columns):
                            n -= 1
                            print '{0:2.0f}'.format(self.propid[n]),
                        print

        elif _DEBUG:
            print '  event time is',event.time,'but update time is', \
                  self.next_update[event.link],'so event will be ignored'
                  
    
    def update_component_data(self, new_node_state_array):
        """
        Call this method to update all data held by the component, if, for
        example, another component or boundary conditions modify the node 
        statuses outside the component between run steps.
        
        This method updates all necessary properties, including both node and
        link states.
        
        *new_node_state_array* is the updated list of node states, which must
        still all be compatible with the state list originally supplied to
        this component.
        """
        self.set_node_state_grid(new_node_state_array)
        self.assign_link_states_from_node_types()
        self.push_transitions_to_event_queue()
        
                  
    def run(self, run_duration, node_state_grid=None, plot_each_transition=False,
            plotter=None):
        """
        Runs the model forward for a specified period of time.
        
        Parameters
        ----------
        run_duration : float
        	Length of time to run
        node_state_grid : 1D array of ints (x number of nodes) (optional)
        	Node states (if given, replaces model's current node state grid)
        plot_each_transition : bool (optional)
        	Option to display the grid after each transition
        plotter : CAPlotter object (optional)
        	Needed if caller wants to plot after every transition
        """
        if node_state_grid is not None:
            self.set_node_state_grid(node_state_grid)
    
        # Continue until we've run out of either time or events
        while self.current_time < run_duration and self.event_queue:
        
            if _DEBUG:
                print 'Current Time = ', self.current_time
        
            # Pick the next transition event from the event queue
            ev = heappop(self.event_queue)
        
            if _DEBUG:
                print 'Event:',ev.time,ev.link,ev.xn_to
        
            self.do_transition(ev, self.current_time, plot_each_transition, plotter)
            
            # Update current time
            self.current_time = ev.time

        
def example_test2():
    
    #from landlab.io.netcdf import write_netcdf
    
    # INITIALIZE

    # User-defined parameters
    nr = 128
    nc = 128
    plot_interval = 0.5
    #next_plot = plot_interval
    run_duration = 50.0
    report_interval = 5.0  # report interval, in real-time seconds
    
    # Initialize real time
    current_real_time = time.time()
    next_report = current_real_time + report_interval

    # Create grid and set up boundaries
    mg = RasterModelGrid(nr, nc, 1.0)
    #mg.set_inactive_boundaries(True, True, True, True)
    
    # Transition data here represent a body of fractured rock, with rock 
    # represented by nodes with state 0, and saprolite (weathered rock)
    # represented by nodes with state 1. Node pairs (links) with 0-1 or 1-0
    # can undergo a transition to 1-1, representing chemical weathering of the
    # rock.
    ns_dict = { 0 : 'air', 1 : 'immobile soil', 2 : 'mobile soil' }
    xn_list = setup_transition_list2()

    # The initial grid represents a domain with half immobile soil, half air
    node_state_grid = mg.add_zeros('node', 'node_state_map', dtype=int)
    if _DEBUG:
        print (numpy.where(mg.node_y<nr/2),)
    (lower_half,) = numpy.where(mg.node_y<nr/2)
    node_state_grid[lower_half] = 1
    
    # Set the left and right boundary conditions
    node_state_grid[mg.left_edge_node_ids()] = 0
    node_state_grid[mg.right_edge_node_ids()] = 0
    
    # Create the CA model
    ca = LandlabCellularAutomaton(mg, ns_dict, xn_list, node_state_grid)
    
    print 'INITIALIZING'
    n = ca.grid.number_of_nodes
    if _DEBUG:
        for r in range(ca.grid.number_of_node_rows):
            for c in range(ca.grid.number_of_node_columns):
                n -= 1
                print '{0:.0f}'.format(ca.node_state[n]),
            print
        
    ca_plotter = CAPlotter(ca)
    
    # RUN
    current_time = 0.0
    #time_slice =  0
    #filename = 'soil_ca1-'+str(time_slice).zfill(5)+'.nc'
    #write_netcdf(filename, ca.grid)
    while current_time < run_duration:
        
        # Once in a while, print out simulation and real time to let the user
        # know that the sim is running ok
        current_real_time = time.time()
        if current_real_time >= next_report:
            print 'Current sim time',current_time,'(',100*current_time/run_duration,'%)'
            next_report = current_real_time + report_interval
        
        ca.run(current_time+plot_interval, ca.node_state, 
               plot_each_transition=False) #, plotter=ca_plotter)
        current_time += plot_interval
        if _DEBUG:
            print 'time:',current_time
            print 'ca time:',ca.current_time
        ca_plotter.update_plot()
        #time_slice += 1
        #filename = 'soil_ca1-'+str(time_slice).zfill(5)+'.nc'
        #write_netcdf(filename, ca.grid)
        if _DEBUG:
            n = ca.grid.number_of_nodes
            for r in range(ca.grid.number_of_node_rows):
                for c in range(ca.grid.number_of_node_columns):
                    n -= 1
                    print '{0:.0f}'.format(ca.node_state[n]),
                print
        
        
    # FINALIZE
    
    # Plot
    ca_plotter.finalize()
        
        
def setup_transition_list():
    """
    Creates and returns a list of Transition() objects. This is a "custom"
    function in the sense that any particular application is determined by the
    transition rules that are created here.
    """
    xn_list = []
    
    xn_list.append( Transition(1, 3, 1., 'weathering') ) # rock-sap to sap-sap
    xn_list.append( Transition(2, 3, 1., 'weathering') ) # sap-rock to sap-sap
        
    if False and _DEBUG:
        print
        print 'setup_transition_list(): list has',len(xn_list),'transitions:'
        for t in xn_list:
            print '  From state',t.from_state,'to state',t.to_state,'at rate',t.rate,'called',t.name
        
    return xn_list
    
    
def setup_transition_list2():
    """
    This one is a crude model of "sticky" particles on a gravity-free hill!
    
    There are three cell states: air (A=0), immobile soil (S=1), and mobile
    soil (M=2)
    
    The transition rules are:
        
        1. Mobilization: A-S > A-M (all orientations equal)
        2. Motion: A-M > M-A (")
        3. Sticking: A-M > A-S (")

    Example of link states associated with 3 node states (0, 1, and 2):
    
    Link states with "horizontal" orientation:

    State:    Pair:  
    0            0-0
    1            0-1
    2            0-2
    3            1-0
    4            1-1
    5            1-2
    6            2-0
    7            2-1
    8            2-2

    Link states with "vertical" orientation (the first number in each pair is
    the lower of the two, i.e., the one with the smaller y coordinate):
    
    State:    Pair:  
    9            0^0
    10           0^1
    11           0^2
    12           1^0
    13           1^1
    14           1^2
    15           2^0
    16           2^1
    17           2^2
   
    """
    xn_list = []
    
    xn_list.append( Transition(1, 2, 1., 'mobilization') ) # 
    xn_list.append( Transition(3, 6, 1., 'mobilization') ) # 
    xn_list.append( Transition(10, 11, 1., 'mobilization') ) # 
    xn_list.append( Transition(12, 15, 1., 'mobilization') ) # 
    xn_list.append( Transition(2, 6, 10., 'left motion') )
    xn_list.append( Transition(6, 2, 10., 'right motion') )
    xn_list.append( Transition(11, 15, 1000., 'downward motion') )
    xn_list.append( Transition(15, 11, 1.0, 'upward motion') )    
        
    if False and _DEBUG:
        print
        print 'setup_transition_list2(): list has',len(xn_list),'transitions:'
        for t in xn_list:
            print '  From state',t.from_state,'to state',t.to_state,'at rate',t.rate,'called',t.name
        
    return xn_list
    
    
if __name__ == "__main__":
    import doctest
    doctest.testmod()
    #main()



import numpy as np


from . import nodes
from ..base import CORE_NODE, CLOSED_BOUNDARY
from ..unstructured.links import LinkGrid


def shape_of_vertical_links(shape):
    """Shape of vertical link grid.

    Number of rows and columns of *vertical* links that connect nodes in a
    structured grid of quadrilaterals.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    tuple of int :
        Shape of the vertical links in grid.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import shape_of_vertical_links
    >>> shape_of_vertical_links((3, 4))
    (2, 4)
    """
    return (shape[0] - 1,  shape[1])


def shape_of_horizontal_links(shape):
    """Shape of horizontal link grid.

    Number of rows and columns of *horizontal* links that connect nodes in a
    structured grid of quadrilaterals.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    tuple of int :
        Shape of the horizontal links in grid.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import shape_of_horizontal_links
    >>> shape_of_horizontal_links((3, 4))
    (3, 3)
    """
    return (shape[0],  shape[1] - 1)


def number_of_vertical_links(shape):
    """Number of vertical links in a structured quad grid.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    int :
        Number of vertical links in grid.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import number_of_vertical_links
    >>> number_of_vertical_links((3, 4))
    8
    """
    return np.prod(shape_of_vertical_links(shape))


def number_of_horizontal_links(shape):
    """Number of horizontal links in a structured quad grid.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    int :
        Number of horizontal links in grid.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import number_of_horizontal_links
    >>> number_of_horizontal_links((3, 4))
    9
    """
    return np.prod(shape_of_horizontal_links(shape))


def number_of_links(shape):
    """Number of links in a structured quad grid.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    int :
        Number of links in grid.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import number_of_links
    >>> number_of_links((3, 4))
    17
    """
    return number_of_vertical_links(shape) + number_of_horizontal_links(shape)


def vertical_link_ids(shape):
    """IDs of vertical links in a structured quad grid.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    (M, N) ndarray :
        Array of link IDs.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import vertical_link_ids
    >>> vertical_link_ids((3, 4))
    array([[0, 1, 2, 3],
           [4, 5, 6, 7]])
    """
    link_ids = np.arange(number_of_vertical_links(shape))
    return link_ids.reshape(shape_of_vertical_links(shape))


def horizontal_link_ids(shape):
    """IDs of horizontal links in a structured quad grid.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    (M, N) ndarray :
        Array of link IDs.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import horizontal_link_ids
    >>> horizontal_link_ids((3, 4))
    array([[ 8,  9, 10],
           [11, 12, 13],
           [14, 15, 16]])
    """
    link_ids = (np.arange(number_of_horizontal_links(shape)) +
                number_of_vertical_links(shape))
    return link_ids.reshape(shape_of_horizontal_links(shape))


def number_of_links_per_node(shape):
    """Number of links touching each node.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    ndarray :
        Array of number of links per node.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import (number_of_links_per_node,
    ...                                            number_of_in_links_per_node,
    ...                                            number_of_out_links_per_node)
    >>> number_of_links_per_node((3, 4))
    array([[2, 3, 3, 2],
           [3, 4, 4, 3],
           [2, 3, 3, 2]])
    >>> (number_of_in_links_per_node((3, 4)) +
    ...  number_of_out_links_per_node((3, 4)))
    array([[2, 3, 3, 2],
           [3, 4, 4, 3],
           [2, 3, 3, 2]])
    """
    link_count = np.empty(shape, np.int)
    link_count[1:-1, 1:-1] = 4
    link_count[(0, -1), 1:-1] = 3
    link_count[1:-1, (0, -1)] = 3
    link_count[(0, 0, -1, -1), (0, -1, 0, -1)] = 2
    return link_count


def number_of_in_links_per_node(shape):
    """Number of links entering each node.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    ndarray :
        Array of number of in-links per node.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import number_of_in_links_per_node
    >>> number_of_in_links_per_node((3, 4))
    array([[0, 1, 1, 1],
           [1, 2, 2, 2],
           [1, 2, 2, 2]])
    """
    link_count = np.empty(shape, np.int)
    link_count[1:, 1:] = 2
    link_count[0, 0] = 0
    link_count[0, 1:] = 1
    link_count[1:, 0] = 1
    return link_count


def number_of_out_links_per_node(shape):
    """Number of links leaving each node.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    ndarray :
        Array of number of out-links per node.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import number_of_out_links_per_node
    >>> number_of_out_links_per_node((3, 4))
    array([[2, 2, 2, 1],
           [2, 2, 2, 1],
           [1, 1, 1, 0]])
    """
    link_count = np.empty(shape, np.int)
    link_count[:-1, :-1] = 2
    link_count[-1, -1] = 0
    link_count[-1, :-1] = 1
    link_count[:-1, -1] = 1
    return link_count


def _node_out_link_ids(shape):
    """Link IDs for links leaving each node.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    tuple :
        Tuple of array of link IDs as (vertical_links, horizontal_links).

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import _node_out_link_ids
    >>> (vert, horiz) = _node_out_link_ids((3, 4))
    >>> vert
    array([[ 0,  1,  2,  3],
           [ 4,  5,  6,  7],
           [-1, -1, -1, -1]])
    >>> horiz
    array([[ 8,  9, 10, -1],
           [11, 12, 13, -1],
           [14, 15, 16, -1]])
    """
    node_horizontal_link_ids = np.empty(shape, np.int)
    node_horizontal_link_ids[:, :-1] = horizontal_link_ids(shape)
    node_horizontal_link_ids[:, -1] = -1

    node_vertical_link_ids = np.empty(shape, np.int)
    node_vertical_link_ids[:-1, :] = vertical_link_ids(shape)
    node_vertical_link_ids[-1, :] = -1

    return node_vertical_link_ids, node_horizontal_link_ids


def _node_in_link_ids(shape):
    """Link IDs for links entering each node.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    tuple :
        Tuple of array of link IDs as (vertical_links, horizontal_links).

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import _node_in_link_ids
    >>> (vert, horiz) = _node_in_link_ids((3, 4))
    >>> vert
    array([[-1, -1, -1, -1],
           [ 0,  1,  2,  3],
           [ 4,  5,  6,  7]])
    >>> horiz
    array([[-1,  8,  9, 10],
           [-1, 11, 12, 13],
           [-1, 14, 15, 16]])
    """
    node_horizontal_link_ids = np.empty(shape, np.int)
    node_horizontal_link_ids[:, 1:] = horizontal_link_ids(shape)
    node_horizontal_link_ids[:, 0] = -1

    node_vertical_link_ids = np.empty(shape, np.int)
    node_vertical_link_ids[1:, :] = vertical_link_ids(shape)
    node_vertical_link_ids[0, :] = -1

    return node_vertical_link_ids, node_horizontal_link_ids


def node_in_link_ids(shape):
    """Link IDs for links entering each node.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    tuple :
        Tuple of array of link IDs as (vertical_links, horizontal_links).

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import node_in_link_ids
    >>> (links, offset) = node_in_link_ids((3, 4))
    >>> links
    array([ 8,  9, 10,  0,  1, 11,  2, 12,  3, 13,  4,  5, 14,  6, 15,  7, 16])
    >>> offset
    array([ 0,  0,  1,  2,  3,  4,  6,  8, 10, 11, 13, 15, 17])

    The links entering the 1st, 5th, and last node,

    >>> for link in [0, 4, 11]: links[offset[link]:offset[link + 1]]
    array([], dtype=int64)
    array([0])
    array([ 7, 16])
    """
    (in_vert, in_horiz) = _node_in_link_ids(shape)
    node_link_ids = np.vstack((in_vert.flat, in_horiz.flat)).T
    #offset = np.cumsum(number_of_in_links_per_node(shape))

    offset = np.empty(nodes.number_of_nodes(shape) + 1, dtype=int)
    np.cumsum(number_of_in_links_per_node(shape), out=offset[1:])
    offset[0] = 0
    return node_link_ids[node_link_ids >= 0], offset


def node_out_link_ids(shape):
    """Link IDs for links leaving each node.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    tuple :
        Tuple of array of link IDs as (vertical_links, horizontal_links).

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import node_out_link_ids
    >>> (links, offset) = node_out_link_ids((3, 4))
    >>> links
    array([ 0,  8,  1,  9,  2, 10,  3,  4, 11,  5, 12,  6, 13,  7, 14, 15, 16])
    >>> offset
    array([ 0,  2,  4,  6,  7,  9, 11, 13, 14, 15, 16, 17, 17])

    The links leaving the 1st, 8th, and last node,

    >>> for link in [0, 7, 11]: links[offset[link]:offset[link + 1]]
    array([0, 8])
    array([7])
    array([], dtype=int64)
    """
    (out_vert, out_horiz) = _node_out_link_ids(shape)
    node_link_ids = np.vstack((out_vert.flat, out_horiz.flat)).T
    offset = np.empty(nodes.number_of_nodes(shape) + 1, dtype=int)
    np.cumsum(number_of_out_links_per_node(shape), out=offset[1:])
    offset[0] = 0
    return node_link_ids[node_link_ids >= 0], offset


def node_link_ids(shape):
    """Link IDs for links entering and leaving each node.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    tuple :
        Tuple of array of link IDs and offsets into link array.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import node_link_ids
    >>> (links, offset) = node_link_ids((3, 4))
    >>> links
    array([ 0,  8,  8,  1,  9,  9,  2, 10, 10,  3,  0,  4, 11,  1, 11,  5, 12,
            2, 12,  6, 13,  3, 13,  7,  4, 14,  5, 14, 15,  6, 15, 16,  7, 16])
    >>> offset
    array([ 0,  2,  5,  8, 10, 13, 17, 21, 24, 26, 29, 32, 34])

    The links attached to node 0

    >>> links[offset[0]:offset[1]]
    array([0, 8])

    The links attached to node 5

    >>> links[offset[5]:offset[6]]
    array([ 1, 11,  5, 12])
    """
    (in_vert, in_horiz) = _node_in_link_ids(shape)
    (out_vert, out_horiz) = _node_out_link_ids(shape)
    node_link_ids = np.vstack((in_vert.flat, in_horiz.flat, out_vert.flat, out_horiz.flat)).T

    offset = np.empty(nodes.number_of_nodes(shape) + 1, dtype=int)
    np.cumsum(number_of_links_per_node(shape), out=offset[1:])
    offset[0] = 0

    return node_link_ids[node_link_ids >= 0], offset


def node_id_at_link_start(shape):
    """Node ID at start of links.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    ndarray :
        Node IDs at start of links.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import node_id_at_link_start
    >>> node_id_at_link_start((3, 4))
    array([ 0,  1,  2,  3,  4,  5,  6,  7,  0,  1,  2,  4,  5,  6,  8,  9, 10])
    """
    all_node_ids = nodes.node_ids(shape)
    return np.concatenate((all_node_ids[:-1, :].flat, all_node_ids[:, :-1].flat))


def node_id_at_link_end(shape):
    """Node ID at end of links.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    ndarray :
        Node IDs at end of links.

    Examples
    --------
    >>> from landlab.grid.structured_quad.links import node_id_at_link_end
    >>> node_id_at_link_end((3, 4))
    array([ 4,  5,  6,  7,  8,  9, 10, 11,  1,  2,  3,  5,  6,  7,  9, 10, 11])
    """
    all_node_ids = nodes.node_ids(shape)
    return np.concatenate((all_node_ids[1:, :].flat, all_node_ids[:, 1:].flat))


def is_active_link(shape, node_status):
    """IDs of active links.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    node_status : array_link
        Status of nodes in grid.

    Returns
    -------
    ndarray :
        Links IDs at the active links.

    Examples
    --------
    >>> from landlab.grid.structured_quad.nodes import status_with_perimeter_as_boundary
    >>> from landlab.grid.structured_quad.links import is_active_link
    >>> status = status_with_perimeter_as_boundary((3, 4))
    >>> is_active_link((3, 4), status)
    array([False,  True,  True, False, False,  True,  True, False, False,
           False, False,  True,  True,  True, False, False, False], dtype=bool)
    """
    if np.prod(shape) != node_status.size:
        raise ValueError('node status array does not match size of grid '
                         '(%d != %d)' % (np.prod(shape), len(node_status)))

    status_at_link_start = node_status.flat[node_id_at_link_start(shape)]
    status_at_link_end = node_status.flat[node_id_at_link_end(shape)]

    return (((status_at_link_start == CORE_NODE) &
             ~ (status_at_link_end == CLOSED_BOUNDARY)) |
            ((status_at_link_end == CORE_NODE) &
             ~ (status_at_link_start == CLOSED_BOUNDARY)))


def active_link_ids(shape, node_status):
    """IDs of active links.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    node_status : array_link
        Status of nodes in grid.

    Returns
    -------
    ndarray :
        Links IDs at the active links.

    Examples
    --------
    >>> from landlab.grid.structured_quad.nodes import status_with_perimeter_as_boundary
    >>> from landlab.grid.structured_quad.links import active_link_ids
    >>> status = status_with_perimeter_as_boundary((3, 4))
    >>> active_link_ids((3, 4), status)
    array([ 1,  2,  5,  6, 11, 12, 13])
    """
    return np.where(is_active_link(shape, node_status))[0]

    
def horizontal_active_link_ids(shape, active_link_ids, BAD_INDEX_VALUE=-1):
    """Get IDs of horizontal active links.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    active_link_ids : array of int
        Array of all active link ids
    BAD_INDEX_VALUE: int, optional
        Value assigned to inactive indicies in the array.

    Returns
    -------
    ndarray :
        Link IDs at the HORIZONTAL active links. Length of number_of_horizontal_links.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.grid.structured_quad.links import active_link_ids, horizontal_active_link_ids
    >>>
    >>> rmg = RasterModelGrid(4, 5)
    >>> rmg.set_closed_boundaries_at_grid_edges(True, True, True, True)
    >>> status = rmg.node_status 
    >>> 
    >>> active_link_ids = active_link_ids((4,5), status)
    >>> horizontal_active_link_ids((4,5), active_link_ids)
    array([-1, -1, -1, -1, -1, 20, 21, -1, -1, 24, 25, -1, -1, -1, -1, -1])

    

    Example grid: Indicies are given for active horizontal links in the 4x5 grid space.
                "X" indicates inactive links, "V" indicates vertical ids

    ::

          *---X-->*---X-->*---X-->*---X-->*
          ^       ^       ^       ^       ^   
          X       X       X       X       X      
          |       |       |       |       |         
          *---X-->*--24-->*--25-->*---X-->*
          ^       ^       ^       ^       ^       
          X       V       V       V       X       
          |       |       |       |       |       
          *---X-->*--20-->*--21-->*---X-->*
          ^       ^       ^       ^       ^   
          X       X       X       X       X       
          |       |       |       |       |       
          *---X-->*---X-->*---X-->*---X-->*
    """
    # For horizontal links, we need to start with a list of '-1' indices with 
    # length of number_of_links
    horizontal_links = np.ones(number_of_links(shape))*BAD_INDEX_VALUE
    
    # We will need the number of rows and columns from input argument 'shape'
    rows = shape[0]
    cols = shape[1]
    
    # In a structured quad, the minimum horizontal link id is equal to the 
    # number of columns * (number of rows - 1)
    min_hori_id = cols*(rows-1)
    
    # We will use list comprehension to get *just* the horizontal link ids
    # from the active_link_ids input argument.
    horizontal_ids = [i for i in active_link_ids if i>= min_hori_id]
    
    # In the array of '-1' we input the horizontal active link ids
    horizontal_links[horizontal_ids] = horizontal_ids
    
    # To get an array of len number_of_horizontal_links, we need to clip off the 
    # number of vertical links. We do this by starting at the "minimum horizontal
    # link id" found above and going to the end of the list. 
    horizontal_links = horizontal_links[min_hori_id:]
    horizontal_links = horizontal_links.astype(int)
    
    # Return an array with length of number_of_vertical_links that has '-1' for
    # inactive links and the active link id for active links
    return horizontal_links


def vertical_active_link_ids(shape, active_link_ids, BAD_INDEX_VALUE=-1):
    """Get IDs of vertical active links.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    active_link_ids : array of int
        Array of all active link ids
    BAD_INDEX_VALUE: int, optional
        Value assigned to inactive indicies in the array.

    Returns
    -------
    ndarray :
        Link IDs at the VERTICAL active links. Length of number_of_vertical_links.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.grid.structured_quad.links import active_link_ids, vertical_active_link_ids
    >>>
    >>> rmg = RasterModelGrid(4, 5)
    >>> rmg.set_closed_boundaries_at_grid_edges(True, True, True, True)
    >>> status = rmg.node_status 
    >>> 
    >>> active_link_ids = active_link_ids((4,5), status)
    >>> vertical_active_link_ids((4,5), active_link_ids)
    array([-1, -1, -1, -1, -1, -1,  6,  7,  8, -1, -1, -1, -1, -1, -1])
    
    
    Example grid: Indicies are given for active vertical links in the 4x5 grid space.
                "X" indicates inactive links, "H" indicates horizontal ids
                
    ::

          *---X-->*---X-->*---X-->*---X-->*
          ^       ^       ^       ^       ^   
          X       X       X       X       X      
          |       |       |       |       |         
          *---X-->*---H-->*---H-->*---X-->*
          ^       ^       ^       ^       ^       
          X       6       7       8       X       
          |       |       |       |       |       
          *---X-->*---H-->*---H-->*---X-->*
          ^       ^       ^       ^       ^   
          X       X       X       X       X       
          |       |       |       |       |       
          *---X-->*---X-->*---X-->*---X-->*
    """
    # Set up an array of '-1' indices with length of number_of_vertical_links
    vertical_links = np.ones(number_of_vertical_links(shape))*BAD_INDEX_VALUE
    
    # We will need the number of rows and columns from input argument 'shape'
    rows = shape[0]
    cols = shape[1]
    
    # In a structured quad, the maximum vertical link id is one less than the 
    # number of columns * (number of rows - 1)
    max_vert_id = cols*(rows-1)
    
    # We will use list comprehension to get *just* the vertical link ids
    # from the active_link_ids input argument.
    vertical_ids = [i for i in active_link_ids if i < max_vert_id]
    
    # In the array of '-1's, we input the active link ids. 
    vertical_links[vertical_ids] = vertical_ids
    vertical_links = vertical_links.astype(int)

    # Return an array with length of number_of_vertical_links that has '-1' for
    # inactive links and the active link id for active links
    return vertical_links
    
def find_horizontal_south_neighbor(shape, horizontal_link_ids, BAD_INDEX_VALUE=-1):
    """Get IDs of SOUTH, horizontal link neighbor

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    horizontal_active_ids : array of int
        Array of all horizontal link ids - MUST BE ARRAY OF LEN(HORIZONTAL_LINKS)
    BAD_INDEX_VALUE: int, optional
        Value assigned to inactive indicies in the array.

    Returns
    -------
    ndarray :
        Link IDs of south horizontal neighbor active links. Length of number_of_horizontal_links.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.grid.structured_quad.links import active_link_ids, horizontal_active_link_ids, find_horizontal_south_neighbor

    >>> rmg = RasterModelGrid(4, 5)

    >>> active_link_ids = active_link_ids((4,5), rmg.node_status)
    >>> horizontal_active_ids = horizontal_active_link_ids((4,5), active_link_ids)
    >>> find_horizontal_south_neighbor((4,5), horizontal_active_ids)
    array([-1, -1, -1, -1, -1, -1, -1, -1, 19, 20, 21, 22, 23, 24, 25, 26])

    
    Example grid: Indicies are given for active horizontal links in the 4x5 grid space.
                  Only horizontal links are shown, asterisks (*) represent nodes.
                  In this example, link 24 has one south neighbor, link 20. In the 
                  find_horizontal_south_neighbor array, the index of 20 (the south link)
                  corresponds with link 24. Similarly, for link 25, the south neighbor
                  returned is link 21, etc. When no active link exists as a south neighbor
                  (in the case of link 19), BAD_INDEX_VALUE is returned. 
                  
    ::

          *------>*------>*------>*------>*
       
       
            
          *--23-->*--24-->*--25-->*--26-->*
    
    
    
          *--19-->*--20-->*--21-->*--22-->*


              
          *------>*------>*------>*------>*
    """

    # First, we find the shape of the horizontal link array given the shape
    # of the raster model grid. In our example, the shape of horizontal links for
    # a grid of 4 rows and 5 columns is 4 rows of horizontal links and 4 columns of
    # horizontal links.
    horizontal_2d_shape = shape_of_horizontal_links(shape)
    
    # Then, we reshape the flattend (1-D) horizontal_link_id array into the shape
    # provided by the shape_of_horizontal_links() function.
    horizontal_2d_array = np.reshape(horizontal_link_ids, horizontal_2d_shape)
    
    # To find south links, we need to shift the IDs in the 2-D array. We first
    # insert a row of BAD_INDEX_VALUE into the top row of the array
    horizontal_link_ids = np.insert(horizontal_2d_array, [0], BAD_INDEX_VALUE, axis=0)
    
    # We find the updated array shape and number of rows for the updated array. 
    row_len = np.shape(horizontal_link_ids)[0]
    
    # To get back to the correct array size (the one found using shape_of_horizontal_links),
    # we delete the last row in the 2-D array
    link_ids  = np.delete(horizontal_link_ids, [row_len-1], axis=0)
    
    # Once we have shifted the 2-D array and removed extra indices, we can flatten
    # the output array to a 1-D array with length of number_of_horizontal_links.
    south_horizontal_neighbors = link_ids.flatten()
    
    return south_horizontal_neighbors   
    
def find_horizontal_west_neighbor(shape, horizontal_link_ids, BAD_INDEX_VALUE=-1):
    """Get IDs of west, horizontal link neighbor

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    horizontal_active_ids : array of int
        Array of all horizontal link ids - MUST BE ARRAY OF LEN(HORIZONTAL_LINKS)
    BAD_INDEX_VALUE: int, optional
        Value assigned to inactive indicies in the array.

    Returns
    -------
    ndarray :
        Link IDs of west horizontal neighbor active links. Length of number_of_horizontal_links.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.grid.structured_quad.links import active_link_ids, horizontal_active_link_ids, find_horizontal_west_neighbor

    >>> rmg = RasterModelGrid(4, 5)
    >>> rmg.set_closed_boundaries_at_grid_edges(True, True, True, True)
    >>> status = rmg.node_status 

    >>> active_link_ids = active_link_ids((4,5), status)
    >>> horizontal_active_ids = horizontal_active_link_ids((4,5), active_link_ids)
    >>> find_horizontal_west_neighbor((4,5), horizontal_active_ids)
    array([-1, -1, -1, -1, -1, -1, 20, 21, -1, -1, 24, 25, -1, -1, -1, -1])

    
    Example grid: Indicies are given for active horizontal links in the 4x5 grid space.
                  Only horizontal links are shown, asterisks (*) represent nodes.
                  In this example, link 21 has one west neighbor, link 20. In the 
                  find_horizontal_west_neighbor array, the index of 20 (the west link)
                  corresponds with link 21. Similarly, for link 25, the west neighbor
                  returned is link 24, etc. When no active link exists as a west neighbor
                  (in the case of link 20), BAD_INDEX_VALUE is returned. 

    ::

          *------>*------>*------>*------>*
       
       
            
          *------>*--24-->*--25--->*----->*
    
    
    
          *------>*--20-->*--21-->*------>*


              
          *------>*------>*------>*------>*
    """
    
    # First, we find the shape of the horizontal link array given the shape
    # of the raster model grid. In our example, the shape of horizontal links for
    # a grid of 4 rows and 5 columns is 4 rows of horizontal links and 4 columns of
    # horizontal links. 
    horizontal_2d_shape = shape_of_horizontal_links(shape)
    
    # Then, we reshape the flattend (1-D) horizontal_link_id array into the shape
    # provided by the shape_of_horizontal_links() function. 
    horizontal_2d_array = np.reshape(horizontal_link_ids, horizontal_2d_shape)
     
    # To find west links, we need to shift the IDs in the 2-D array. We insert
    # a column of BAD_INDEX_VALUE into the first column of the array. 
    horizontal_link_ids = np.insert(horizontal_2d_array, [0], BAD_INDEX_VALUE, axis=1)
    
    # We find the updated array shape and number of columns for the updated array. 
    row_len = np.shape(horizontal_link_ids)[1]
    
    # To get back to the correct array size (the one found using shape_of_horizontal_links),
    # we delete the very LAST column of the 2-D array. (Any link final column in the 2-D array
    # cannot be a western neighbor anyway). 
    horizontal_link_ids = np.delete(horizontal_link_ids, [row_len-1], axis=1)
    
    # Once we have shifted the 2-D array and removed extra indices, we can flatten
    # the output array to a 1-D array with length of number_of_horizontal_links.
    west_horizontal_neighbors = horizontal_link_ids.flatten()
    
    return west_horizontal_neighbors


def find_horizontal_north_neighbor(shape, horizontal_link_ids, BAD_INDEX_VALUE=-1):
    """Get IDs of NORTH, horizontal link neighbor

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    horizontal_active_ids : array of int
        Array of all horizontal link ids - MUST BE ARRAY OF LEN(HORIZONTAL_LINKS)
    BAD_INDEX_VALUE: int, optional
        Value assigned to inactive indicies in the array.

    Returns
    -------
    ndarray :
        Link IDs of north horizontal neighbor links. Length of number_of_horizontal_links.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.grid.structured_quad.links import active_link_ids, horizontal_active_link_ids, find_horizontal_north_neighbor

    >>> rmg = RasterModelGrid(4, 5)

    >>> active_link_ids = active_link_ids((4,5), rmg.node_status)
    >>> horizontal_active_ids = horizontal_active_link_ids((4,5), active_link_ids)
    >>> find_horizontal_north_neighbor((4,5), horizontal_active_ids)
    array([19, 20, 21, 22, 23, 24, 25, 26, -1, -1, -1, -1, -1, -1, -1, -1])

    Example grid: Indicies are given for active horizontal links in the 4x5 grid space.
                  Only horizontal links are shown, asterisks (*) represent nodes.
                  In this example, link 20 has one north neighbor, link 24. In the 
                  find_horizontal_north_neighbor array, the index of 24 (the north link)
                  corresponds with link 20. Similarly, for link 21, the north neighbor
                  returned is link 25, etc. When no active link exists as a north neighbor
                  (in the case of link 23), BAD_INDEX_VALUE is returned. 

    ::

          *------>*------>*------>*------>*
       
       
            
          *--23-->*--24-->*--25-->*--26-->*
    
    
    
          *--19-->*--20-->*--21-->*--22-->*


              
          *------>*------>*------>*------>*
    """

    # First, we find the shape of the horizontal link array given the shape
    # of the raster model grid. In our example, the shape of horizontal links for
    # a grid of 4 rows and 5 columns is 4 rows of horizontal links and 4 columns of
    # horizontal links.
    horizontal_2d_shape = shape_of_horizontal_links(shape)
    
    # Then, we reshape the flattend (1-D) horizontal_link_id array into the shape
    # provided by the shape_of_horizontal_links() function.
    horizontal_2d_array = np.reshape(horizontal_link_ids, horizontal_2d_shape)
    
    # To find north links, we need to shift the IDs in the 2-D array. We first
    # delete the top row of the array
    horizontal_link_ids = np.delete(horizontal_2d_array, [0], axis=0)
    
    # We find the updated array shape and number of rows for the updated array. 
    row_len = np.shape(horizontal_link_ids)[0]
    
    # To get back to the correct array size (the one found using shape_of_horizontal_links),
    # we insert a row (populated with BAD_INDEX_VALUE_ into the end of the 2-D array.
    link_ids  = np.insert(horizontal_link_ids, [row_len], BAD_INDEX_VALUE, axis=0)
    
    # Once we have shifted the 2-D array and removed extra indices, we can flatten
    # the output array to a 1-D array with length of number_of_horizontal_links.
    north_horizontal_neighbors = link_ids.flatten()
    
    return north_horizontal_neighbors   


def find_horizontal_east_neighbor(shape, horizontal_link_ids, BAD_INDEX_VALUE=-1):
    """Get IDs of east, horizontal link neighbor

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    horizontal_active_ids : array of int
        Array of all horizontal link ids - MUST BE ARRAY OF LEN(HORIZONTAL_LINKS)
    BAD_INDEX_VALUE: int, optional
        Value assigned to inactive indicies in the array.

    Returns
    -------
    ndarray :
        Link IDs of east horizontal neighbor active links. Length of number_of_horizontal_links.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.grid.structured_quad.links import active_link_ids, horizontal_active_link_ids, find_horizontal_east_neighbor
    >>>
    >>> rmg = RasterModelGrid(4, 5)
    >>> rmg.set_closed_boundaries_at_grid_edges(True, True, True, True)
    >>> status = rmg.node_status 
    >>> 
    >>> active_link_ids = active_link_ids((4,5), status)
    >>> horizontal_active_ids = horizontal_active_link_ids((4,5), active_link_ids)
    >>> find_horizontal_east_neighbor((4,5), horizontal_active_ids)
    array([-1, -1, -1, -1, 20, 21, -1, -1, 24, 25, -1, -1, -1, -1, -1, -1])

    
    Example grid: Indicies are given for active horizontal links in the 4x5 grid space.
                  Only horizontal links are shown, asterisks (*) represent nodes.
                  In this example, link 20 has one east neighbor, link 21. In the 
                  find_horizontal_east_neighbor array, the index of 21 (the east link)
                  corresponds with link 20. Similarly, for link 24, the east neighbor
                  returned is link 25, etc. When no active link exists as a east neighbor
                  (in the case of link 21), BAD_INDEX_VALUE is returned. 

    ::

          *------>*------>*------>*------>*
       
       
            
          *------>*--24-->*--25--->*----->*
    
    
    
          *------>*--20-->*--21-->*------>*


              
          *------>*------>*------>*------>*
    """
    
    # First, we find the shape of the horizontal link array given the shape
    # of the raster model grid. In our example, the shape of horizontal links for
    # a grid of 4 rows and 5 columns is 4 rows of horizontal links and 4 columns of
    # horizontal links.
    horizontal_2d_shape = shape_of_horizontal_links(shape)
    
    # Then, we reshape the flattend (1-D) horizontal_link_id array into the shape
    # provided by the shape_of_horizontal_links() function.
    horizontal_2d_array = np.reshape(horizontal_link_ids, horizontal_2d_shape)
    
    
    # To find west links, we need to shift the IDs in the 2-D array. We first
    # delete the first column of the array (these values can never be east neighbors
    # anyway.) 
    horizontal_link_ids = np.delete(horizontal_2d_array, [0], axis=1)
    
    # We find the updated array shape and number of columns for the updated array. 
    row_len = np.shape(horizontal_link_ids)[1]
    
    # To get back to the correct array size (the one found using shape_of_horizontal_links),
    # we insert a column of BAD_INDEX_VALUE into the last column spot in the 2-D array.
    link_ids  = np.insert(horizontal_link_ids, [row_len], BAD_INDEX_VALUE, axis=1)
    
    # Once we have shifted the 2-D array and removed extra indices, we can flatten
    # the output array to a 1-D array with length of number_of_horizontal_links.
    east_horizontal_neighbors = link_ids.flatten()
    
    return east_horizontal_neighbors   
               
               
def find_d4_horizontal_neighbors(shape, horizontal_ids, BAD_INDEX_VALUE=-1):
    """Give IDs of all 4 horizontal link neighbors. [S,W,N,E]

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    horizontal_active_ids : array of int
        Array of all horizontal link ids - MUST BE ARRAY OF LEN(HORIZONTAL_LINKS)
    BAD_INDEX_VALUE: int, optional
        Value assigned to inactive indicies in the array.

    Returns
    -------
    ndarray :
        Array of 4 horizontal link neighbors for a given link ID. Returned in
        [S, W, N, E]. 

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.grid.structured_quad.links import active_link_ids, horizontal_active_link_ids, find_d4_horizontal_neighbors
    >>> rmg = RasterModelGrid(4, 5)
    >>> active_link_ids = active_link_ids((4,5), rmg.node_status)
    >>> horizontal_active_ids = horizontal_active_link_ids((4,5), active_link_ids)
    >>> find_d4_horizontal_neighbors((4,5), horizontal_active_ids)
    array([[-1, -1, 19, -1],  
        [-1, -1, 20, -1],
        [-1, -1, 21, -1],
        [-1, -1, 22, -1],
        [-1, -1, 23, 20],
        [-1, 19, 24, 21],
        [-1, 20, 25, 22],
        [-1, 21, 26, -1],
        [19, -1, -1, 24],
        [20, 23, -1, 25],
        [21, 24, -1, 26],
        [22, 25, -1, -1],
        [23, -1, -1, -1],
        [24, -1, -1, -1],
        [25, -1, -1, -1],
        [26, -1, -1, -1]])

    
    Example grid: Indicies are given for active horizontal links in the 4x5 grid space.
                  Only horizontal links are shown, asterisks (*) represent nodes.
                  
                  In this example, link 20 has three neighbors. (links 19, 21, and 24).
                  This function looks for S, W, N, E neighbors and returns an array of
                  indices for each link. For link 20, the returned array would be
                  [-1, 19, 24, 21]. 

    ::

          *------>*------>*------>*------>*
       
       
            
          *------>*--24-->*------>*------>*
    
    
    
          *--19-->*--20-->*--21-->*------>*


              
          *------>*------>*------>*------>*
    """
    ### First we find *south* neighbors...
    south = find_horizontal_south_neighbor(shape, horizontal_ids, BAD_INDEX_VALUE)
    
    ### Then *west* neighbors...
    west = find_horizontal_west_neighbor(shape, horizontal_ids, BAD_INDEX_VALUE) 

    ### Then *north* neighbors...
    north = find_horizontal_north_neighbor(shape, horizontal_ids, BAD_INDEX_VALUE)

    ### Finally, *east* neighbors...
    east = find_horizontal_east_neighbor(shape, horizontal_ids, BAD_INDEX_VALUE)
   
    ### Combine all 4 neighbor arrays into one large array (4 x len_horizontal_links)
    neighbor_array = np.array([south, west, north, east])
    
    ### Transpose the 4 neighbor arrays into a (len_horizontal_links x 4) array.
    neighbor_array = np.transpose(neighbor_array)
    
    ### Output neighbor array. For each input ID, returns [S,W,N,E]
    return neighbor_array            
       
       
def find_d4_horizontal_neighbors_active(shape, horizontal_ids, BAD_INDEX_VALUE=-1):
    """Give IDs of all 4 horizontal link neighbors. [S,W,N,E]

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    horizontal_active_ids : array of int
        Array of all horizontal link ids - MUST BE ARRAY OF LEN(HORIZONTAL_LINKS)
    BAD_INDEX_VALUE: int, optional
        Value assigned to inactive indicies in the array.

    Returns
    -------
    ndarray :
        Array of 4 horizontal link neighbors for a given link ID. Returned in
        [S, W, N, E]. Returns array for only ACTIVE horizontal links.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.grid.structured_quad.links import active_link_ids, horizontal_active_link_ids, find_d4_horizontal_neighbors_active

    >>> rmg = RasterModelGrid(4, 5)

    >>> active_link_ids = active_link_ids((4,5), rmg.node_status)
    >>> horizontal_active_ids = horizontal_active_link_ids((4,5), active_link_ids)
    >>> find_d4_horizontal_neighbors_active((4,5), horizontal_active_ids)
    array([[-1, -1, 23, 20],
        [-1, 19, 24, 21],
        [-1, 20, 25, 22],
        [-1, 21, 26, -1],
        [19, -1, -1, 24],
        [20, 23, -1, 25],
        [21, 24, -1, 26],
        [22, 25, -1, -1]])


    Example grid: Indicies are given for active horizontal links in the 4x5 grid space.
                  Only horizontal links are shown, asterisks (*) represent nodes.
                  
                  In this example, link 20 has three neighbors. (links 19, 21, and 24).
                  This function looks for S, W, N, E neighbors and returns an array of
                  indices for each link. For link 20, the returned array would be
                  [-1, 19, 24, 21]. 

    ::

          *------>*------>*------>*------>*
       
       
            
          *------>*--24-->*------>*------>*
    
    
    
          *--19-->*--20-->*--21-->*------>*


              
          *------>*------>*------>*------>*
    """
    # To do this we simply call the find_d4_horizontal_neighbors() function 
    # which gives the neighbors for ALL horizontal links in an array, even 
    # inactive links. 
    d4_neigh = find_d4_horizontal_neighbors(shape, horizontal_ids, BAD_INDEX_VALUE)
    
    # Now we will just focus on indices that are ACTIVE...
    active_links = np.where(horizontal_ids != BAD_INDEX_VALUE)
    
    # Clip our initial array into a smaller one with just active neighbors
    neighbor_array = d4_neigh[active_links]
    
    # Output neighbor array. For each input ID, returns [S,W,N,E]
    return neighbor_array                                  


def find_vertical_south_neighbor(shape, vertical_link_ids, BAD_INDEX_VALUE=-1):
    """Link IDs of south, vertical link neighbor

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    vertical_link_ids : array of int
        Array of all vertical link ids - MUST BE ARRAY OF LEN(VERTICAL_LINKS)
    BAD_INDEX_VALUE: int, optional
        Value assigned to inactive indicies in the array.

    Returns
    -------
    ndarray :
        Link IDs of *south* vertical neighbor active links. Length of
        number_of_vertical_links.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.grid.structured_quad.links import active_link_ids, vertical_active_link_ids, find_vertical_south_neighbor

    >>> rmg = RasterModelGrid(4, 5)

    >>> active_link_ids = active_link_ids((4,5), rmg.node_status)
    >>> vertical_active_ids = vertical_active_link_ids((4,5), active_link_ids)
    >>> find_vertical_south_neighbor((4,5), vertical_active_ids)
    array([-1, -1, -1, -1, -1, -1,  1,  2,  3, -1, -1,  6,  7,  8, -1])

    Example grid: Indicies are given for active vertical links in the 4x5 grid space.
                  Only vertical links are shown, asterisks (*) represent nodes.
                  In this example, link 6 has one south neighbor, link 1. In the 
                  find_vertical_north_neighbor array, the index of 1 (the south link)
                  corresponds with link 6. Similarly, for link 13, the south neighbor
                  returned is link 8, etc. When no active link exists as a east neighbor
                  (in the case of link 2), BAD_INDEX_VALUE is returned. 
    ::

          *       *       *       *       *
          ^       ^       ^       ^       ^   
          |       11      12      13      |      
          |       |       |       |       |         
          *       *       *       *       *
          ^       ^       ^       ^       ^       
          |       6       7       8       |       
          |       |       |       |       |       
          *       *       *       *       *
          ^       ^       ^       ^       ^   
          |       1       2       3       |       
          |       |       |       |       |       
          *       *       *       *       *
    """
    # First, we find the shape of the vertical link array given the shape
    # of the raster model grid. In our example, the shape of vertical links for
    # a grid of 4 rows and 5 columns is 3 rows of vertical links and 5 columns of
    # vertical links.   
    vertical_2d_shape = shape_of_vertical_links(shape)
    
    # Then, we reshape the flattend (1-D) vertical_link_id array into the shape
    # provided by the shape_of_vertical_links() function.
    vertical_2d_array = np.reshape(vertical_link_ids, vertical_2d_shape)

    # To find south links, we need to shift the IDs in the 2-D array. We insert
    # a row of BAD_INDEX_VALUE into the top row of the 2-D array
    link_ids  = np.insert(vertical_2d_array, [0],  BAD_INDEX_VALUE, axis=0)
    
    # We find the updated array shape and number of rows for the updated array. 
    row_len = np.shape(link_ids)[0]
    
    # To get back to the correct array size (the one found using shape_of_vertical_links),
    # we delete a the last row of the 2-D array. 
    vertical_link_ids = np.delete(link_ids, [row_len-1], axis=0)

    # Once we have shifted the 2-D array and removed extra indices, we can flatten
    # the output array to a 1-D array with length of number_of_vertical_links.
    south_vertical_neighbors = vertical_link_ids.flatten()
    
    return south_vertical_neighbors      

def find_vertical_west_neighbor(shape, vertical_link_ids, BAD_INDEX_VALUE=-1):
    """Link IDs of west, vertical link neighbor

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    vertical_link_ids : array of int
        Array of all vertical link ids- MUST BE ARRAY OF LEN(VERTICAL_LINKS)
    BAD_INDEX_VALUE: int, optional
        Value assigned to inactive indicies in the array.

    Returns
    -------
    ndarray :
        Link IDs of *west* vertical neighbor active links. Length of number_of_vertical_links.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.grid.structured_quad.links import active_link_ids, vertical_active_link_ids, find_vertical_west_neighbor

    >>> rmg = RasterModelGrid(4, 5)
    >>> rmg.set_closed_boundaries_at_grid_edges(True, True, True, True)
    >>> status = rmg.node_status 

    >>> active_link_ids = active_link_ids((4,5), status)
    >>> vertical_active_ids = vertical_active_link_ids((4,5), active_link_ids)
    >>> find_vertical_west_neighbor((4,5), vertical_active_ids)
    array([-1, -1, -1, -1, -1, -1, -1,  6,  7,  8, -1, -1, -1, -1, -1])


    Example grid: Indicies are given for active vertical links in the 4x5 grid space.
                  Only vertical links are shown, asterisks (*) represent nodes.
                  In this example, link 7 has one west neighbor, link 6. In the 
                  find_vertical_west_neighbor array, the index of 6 (the west link)
                  corresponds with link 7. Similarly, for link 8, the west neighbor
                  returned is link 7, etc. When no active link exists as a west neighbor
                  (in the case of link 6), BAD_INDEX_VALUE is returned. 

    ::

          *       *       *       *       *
          ^       ^       ^       ^       ^   
          |       |       |       |       |      
          |       |       |       |       |         
          *       *       *       *       *
          ^       ^       ^       ^       ^       
          |       6       7       8       |       
          |       |       |       |       |       
          *       *       *       *       *
          ^       ^       ^       ^       ^   
          |       |       |       |       |       
          |       |       |       |       |       
          *       *       *       *       *
    """
    # First, we find the shape of the vertical link array given the shape
    # of the raster model grid. In our example, the shape of vertical links for
    # a grid of 4 rows and 5 columns is 3 rows of vertical links and 5 columns of
    # vertical links. 
    vertical_2d_shape = shape_of_vertical_links(shape)
    
    # Then, we reshape the flattend (1-D) vertical_link_id array into the shape
    # provided by the shape_of_vertical_links() function. 
    vertical_2d_array = np.reshape(vertical_link_ids, vertical_2d_shape)
    
    # To find west links, we need to shift the IDs in the 2-D array. We insert
    # a column of BAD_INDEX_VALUE into the first column of the array. 
    vertical_link_ids = np.insert(vertical_2d_array, [0], BAD_INDEX_VALUE, axis=1)
    
    # We find the updated array shape and number of columns for the updated array. 
    row_len = np.shape(vertical_link_ids)[1]
    
    # To get back to the correct array size (the one found using shape_of_vertical_links),
    # we delete the very LAST column of the 2-D array. (Any link final column in the 2-D array
    # cannot be a western neighbor anyway). 
    vertical_link_ids = np.delete(vertical_link_ids, [row_len-1], axis=1)
    
    # Once we have shifted the 2-D array and removed extra indices, we can flatten
    # the output array to a 1-D array with length of number_of_vertical_links.
    west_vertical_neighbors = vertical_link_ids.flatten()
    
    return west_vertical_neighbors


def find_vertical_north_neighbor(shape, vertical_link_ids, BAD_INDEX_VALUE=-1):
    """Link IDs of north, vertical link neighbor

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    vertical_link_ids : array of int
        Array of all vertical link ids- MUST BE ARRAY OF LEN(VERTICAL_LINKS)
    BAD_INDEX_VALUE: int, optional
        Value assigned to inactive indicies in the array.

    Returns
    -------
    ndarray :
        Link IDs of *north* vertical neighbor active links. Length of number_of_vertical_links.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.grid.structured_quad.links import active_link_ids, vertical_active_link_ids, find_vertical_north_neighbor

    >>> rmg = RasterModelGrid(4, 5)

    >>> active_link_ids = active_link_ids((4,5), rmg.node_status)
    >>> vertical_active_ids = vertical_active_link_ids((4,5), active_link_ids)
    >>> find_vertical_north_neighbor((4,5), vertical_active_ids)
    array([-1,  6,  7,  8, -1, -1, 11, 12, 13, -1, -1, -1, -1, -1, -1])


    Example grid: Indicies are given for active vertical links in the 4x5 grid space.
                  Only vertical links are shown, asterisks (*) represent nodes.
                  In this example, link 1 has one north neighbor, link 6. In the 
                  find_vertical_north_neighbor array, the index of 6 (the north link)
                  corresponds with link 1. Similarly, for link 8, the north neighbor
                  returned is link 13, etc. When no active link exists as a east neighbor
                  (in the case of link 11), BAD_INDEX_VALUE is returned. 

    ::

          *       *       *       *       *
          ^       ^       ^       ^       ^   
          |       11      12      13      |      
          |       |       |       |       |         
          *       *       *       *       *
          ^       ^       ^       ^       ^       
          |       6       7       8       |       
          |       |       |       |       |       
          *       *       *       *       *
          ^       ^       ^       ^       ^   
          |       1       2       3       |       
          |       |       |       |       |       
          *       *       *       *       *
    """
    # First, we find the shape of the vertical link array given the shape
    # of the raster model grid. In our example, the shape of vertical links for
    # a grid of 4 rows and 5 columns is 3 rows of vertical links and 5 columns of
    # vertical links.   
    vertical_2d_shape = shape_of_vertical_links(shape)
    
    # Then, we reshape the flattend (1-D) vertical_link_id array into the shape
    # provided by the shape_of_vertical_links() function.
    vertical_2d_array = np.reshape(vertical_link_ids, vertical_2d_shape)

    # To find north links, we need to shift the IDs in the 2-D array. We first
    # delete the first row of the array.
    vertical_link_ids = np.delete(vertical_2d_array, [0], axis=0)

    # We find the updated array shape and number of rows for the updated array. 
    row_len = np.shape(vertical_link_ids)[0]
    
    # To get back to the correct array size (the one found using shape_of_vertical_links),
    # we insert a row (populated with BAD_INDEX_VALUE) into the end of the 2-D array. 
    link_ids  = np.insert(vertical_link_ids, [row_len],  BAD_INDEX_VALUE, axis=0)
    
    # Once we have shifted the 2-D array and removed extra indices, we can flatten
    # the output array to a 1-D array with length of number_of_vertical_links.
    north_vertical_neighbors = link_ids.flatten()
    
    return north_vertical_neighbors
    
    
def find_vertical_east_neighbor(shape, vertical_link_ids, BAD_INDEX_VALUE=-1):
    """Link IDs of east, vertical link neighbor

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    vertical_link_ids : array of int
        Array of all vertical link ids - MUST BE ARRAY OF LEN(VERTICAL_LINKS)
    BAD_INDEX_VALUE: int, optional
        Value assigned to inactive indicies in the array.

    Returns
    -------
    ndarray :
        Link IDs of *east* vertical neighbor active links. Length of number_of_vertical_links.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.grid.structured_quad.links import active_link_ids, vertical_active_link_ids, find_vertical_east_neighbor

    >>> rmg = RasterModelGrid(4, 5)
    >>> rmg.set_closed_boundaries_at_grid_edges(True, True, True, True)
    >>> status = rmg.node_status 

    >>> active_link_ids = active_link_ids((4,5), status)
    >>> vertical_active_ids = vertical_active_link_ids((4,5), active_link_ids)
    >>> find_vertical_east_neighbor((4,5), vertical_active_ids)
    array([-1, -1, -1, -1, -1,  6,  7,  8, -1, -1, -1, -1, -1, -1, -1])


    Example grid: Indicies are given for active vertical links in the 4x5 grid space.
                  Only vertical links are shown, asterisks (*) represent nodes.
                  In this example, link 6 has one east neighbor, link 7. In the 
                  find_vertical_east_neighbor array, the index of 7 (the east link)
                  corresponds with link 6. Similarly, for link 7, the east neighbor
                  returned is link 8, etc. When no active link exists as a east neighbor
                  (in the case of link 8), BAD_INDEX_VALUE is returned. 
                
    ::

          *       *       *       *       *
          ^       ^       ^       ^       ^   
          |       |       |       |       |      
          |       |       |       |       |         
          *       *       *       *       *
          ^       ^       ^       ^       ^       
          |       6       7       8       |       
          |       |       |       |       |       
          *       *       *       *       *
          ^       ^       ^       ^       ^   
          |       |       |       |       |       
          |       |       |       |       |       
          *       *       *       *       *
    """
    # First, we find the shape of the vertical link array given the shape
    # of the raster model grid. In our example, the shape of vertical links for
    # a grid of 4 rows and 5 columns is 3 rows of vertical links and 5 columns of
    # vertical links. 
    vertical_2d_shape = shape_of_vertical_links(shape)
    
    # Then, we reshape the flattend (1-D) vertical_link_id array into the shape
    # provided by the shape_of_vertical_links() function.
    vertical_2d_array = np.reshape(vertical_link_ids, vertical_2d_shape)
    
    
    # To find east links, we need to shift the IDs in the 2-D array. We first
    # delete the first column of the array.
    vertical_link_ids = np.delete(vertical_2d_array, [0], axis=1)
    
    # We find the updated array shape and number of columns for the updated array. 
    row_len = np.shape(vertical_link_ids)[1]
    
    # To get back to the correct array size (the one found using shape_of_vertical_links),
    # we insert a column (populated with BAD_INDEX_VALUE) into the end of the 2-D array. 
    link_ids  = np.insert(vertical_link_ids, [row_len],  BAD_INDEX_VALUE, axis=1)
    
    # Once we have shifted the 2-D array and removed extra indices, we can flatten
    # the output array to a 1-D array with length of number_of_vertical_links.
    east_vertical_neighbors = link_ids.flatten()
    
    return east_vertical_neighbors    
    
    
def find_d4_vertical_neighbors(shape, vertical_ids, BAD_INDEX_VALUE=-1):
    """Give IDs of all 4 vertical link neighbors. [S,W,N,E]

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    vertical_active_ids : array of int
        Array of all vertical link ids - MUST BE ARRAY OF LEN(VERTICAL_LINKS)
    BAD_INDEX_VALUE: int, optional
        Value assigned to inactive indicies in the array.

    Returns
    -------
    ndarray :
        Array of 4 vertical link neighbors for a given link ID. Returned in
        [S, W, N, E]. 

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.grid.structured_quad.links import active_link_ids, vertical_active_link_ids, find_d4_vertical_neighbors

    >>> rmg = RasterModelGrid(4, 5)

    >>> active_link_ids = active_link_ids((4,5), rmg.node_status)
    >>> vertical_active_ids = vertical_active_link_ids((4,5), active_link_ids)
    >>> find_d4_vertical_neighbors((4,5), vertical_active_ids)
    array([[-1, -1, -1,  1],
        [-1, -1,  6,  2],
        [-1,  1,  7,  3],
        [-1,  2,  8, -1],
        [-1,  3, -1, -1],
        [-1, -1, -1,  6],
        [ 1, -1, 11,  7],
        [ 2,  6, 12,  8],
        [ 3,  7, 13, -1],
        [-1,  8, -1, -1],
        [-1, -1, -1, 11],
        [ 6, -1, -1, 12],
        [ 7, 11, -1, 13],
        [ 8, 12, -1, -1],
        [-1, 13, -1, -1]])
    
    
    Example grid: Indicies are given for active vertical links in the 4x5 grid space.
                  Only vertical links are shown, asterisks (*) represent nodes.
                  
                  In this example, link 7 has four neighbors. (links 2, 6, 12 and 8 in 
                  S, W, N, E order)
                  
                  This function looks for S, W, N, E neighbors and returns an array of
                  indices for each link. For link 8, the returned array would be
                  [3, 7, 13, -1]. 

    ::

          *       *       *       *       *
          ^       ^       ^       ^       ^   
          |       11      12     13       |      
          |       |       |       |       |         
          *       *       *       *       *
          ^       ^       ^       ^       ^       
          |       6       7       8       |       
          |       |       |       |       |       
          *       *       *       *       *
          ^       ^       ^       ^       ^   
          |       1       2       3       |       
          |       |       |       |       |       
          *       *       *       *       *
    """
    south = find_vertical_south_neighbor(shape, vertical_ids, BAD_INDEX_VALUE)
    west = find_vertical_west_neighbor(shape, vertical_ids, BAD_INDEX_VALUE) 
    north = find_vertical_north_neighbor(shape, vertical_ids, BAD_INDEX_VALUE)
    east = find_vertical_east_neighbor(shape, vertical_ids, BAD_INDEX_VALUE)   
    neighbor_array = np.array([south, west, north, east])
    neighbor_array = np.transpose(neighbor_array)
    return neighbor_array            


def find_d4_vertical_neighbors_active(shape, vertical_ids, BAD_INDEX_VALUE=-1):
    """Give IDs of all 4 vertical link neighbors. [S,W,N,E]

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    vertical_active_ids : array of int
        Array of all vertical link ids - MUST BE ARRAY OF LEN(VERTICAL_LINKS)
    BAD_INDEX_VALUE: int, optional
        Value assigned to inactive indicies in the array.

    Returns
    -------
    ndarray :
        Array of 4 vertical link neighbors for a given ACTIVE link ID. Returned in
        [S, W, N, E]. 

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.grid.structured_quad.links import active_link_ids, vertical_active_link_ids, find_d4_vertical_neighbors_active

    >>> rmg = RasterModelGrid(4, 5)

    >>> active_link_ids = active_link_ids((4,5), rmg.node_status)
    >>> vertical_active_ids = vertical_active_link_ids((4,5), active_link_ids)
    >>> find_d4_vertical_neighbors_active((4,5), vertical_active_ids)
    array([[-1, -1,  6,  2],
        [-1,  1,  7,  3],
        [-1,  2,  8, -1],
        [ 1, -1, 11,  7],
        [ 2,  6, 12,  8],
        [ 3,  7, 13, -1],
        [ 6, -1, -1, 12],
        [ 7, 11, -1, 13],
        [ 8, 12, -1, -1]])

    
    Example grid: Indicies are given for active horizontal links in the 4x5 grid space.
                  Only horizontal links are shown, asterisks (*) represent nodes.
                  
                  In this example, link 7 has four neighbors. (links 2, 6, 12 and 8 in 
                  S, W, N, E order)
                  
                  This function looks for S, W, N, E neighbors and returns an array of
                  indices for each link. For link 8, the returned array would be
                  [3, 7, 13, -1]. 

    ::

          *       *       *       *       *
          ^       ^       ^       ^       ^   
          |       11      12     13       |      
          |       |       |       |       |         
          *       *       *       *       *
          ^       ^       ^       ^       ^       
          |       6       7       8       |       
          |       |       |       |       |       
          *       *       *       *       *
          ^       ^       ^       ^       ^   
          |       1       2       3       |       
          |       |       |       |       |       
          *       *       *       *       *
    """
    # To do this we simply call the find_d4_vertical_neighbors() function 
    # which gives the neighbors for ALL vertical links in an array, even 
    # inactive links. 
    d4_all_neighbors = find_d4_vertical_neighbors(shape, vertical_ids, BAD_INDEX_VALUE)

    # Now we will just focus on indices that are ACTIVE...
    active_links = np.where(vertical_ids != BAD_INDEX_VALUE)
    
    # Clip our initial array into a smaller one with just active neighbors
    neighbor_array = d4_all_neighbors[active_links]
    
    # Output neighbor array. For each input ID, returns [S,W,N,E]
    return neighbor_array                
                                     
                                                                                                                                                                                                       
class StructuredQuadLinkGrid(LinkGrid):
    def __init__(self, shape):
        link_ends = (node_id_at_link_start(shape), node_id_at_link_end(shape))
        number_of_nodes = np.prod(shape)
        LinkGrid.__init__(self, link_ends, number_of_nodes)
